/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif

#include "util.h"

#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <pthread.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <stdio.h>
#include <inttypes.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>


#if defined HAVE_BACKTRACE && defined HAVE_BACKTRACE_SYMBOLS
    #include <execinfo.h>
#endif

/*
 * Timeout policy for client connections
 */
#define NOREAD_TIMEOUT        60*3

double get_timestamp(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return (tv.tv_sec + tv.tv_usec*1e-6);
}

static pthread_mutex_t reportMutex;
static int debugLevel = LOG_DEBUG;
static void printMessage(char *buf);

void vdsm_init_report(int debug)
{
    pthread_mutex_init(&reportMutex, NULL);
    debugLevel = debug;
}

void vdsm_report(int errlevel, const char *fmt, ... )
{
    if (errlevel <= debugLevel) {
        char buf[BUFSIZ];
        va_list ap;
        pthread_mutex_lock(&reportMutex);

        strncpy(buf, "vdSM: ", BUFSIZ);
        va_start(ap, fmt);
        vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
        va_end(ap);

        printMessage(buf);

        pthread_mutex_unlock(&reportMutex);
    }
}

void vdsm_report_extraLevel(int errlevel, int maxErrlevel, const char *fmt, ... )
{
    if (errlevel <= maxErrlevel) {
        char buf[BUFSIZ];
        va_list ap;
        pthread_mutex_lock(&reportMutex);

        strncpy(buf, "vdSM: ", BUFSIZ);
        va_start(ap, fmt);
        vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
        va_end(ap);

        printMessage(buf);

        pthread_mutex_unlock(&reportMutex);
    }
}

static void printMessage(char *buf)
{
    char buf2[BUFSIZ], *sp, tsbuf[30];
    struct timeval t;
    gettimeofday(&t, NULL);
    strftime(tsbuf, sizeof(tsbuf), "%Y-%m-%d %H:%M:%S", localtime(&t.tv_sec));
    (void)snprintf(buf2, BUFSIZ, "[%s.%03d] ", tsbuf, (int)t.tv_usec / 1000);
    for (sp = buf; *sp != '\0'; sp++)
        if (isprint(*sp) || (isspace(*sp) && (sp[1]=='\0' || sp[2]=='\0')))
            (void)snprintf(buf2+strlen(buf2), 2, "%c", *sp);
        else
            (void)snprintf(buf2+strlen(buf2), 6, "\\x%02x", (unsigned)*sp);
    (void)fputs(buf2, stderr);
}

static const char* default_vdc_port = "8440";

bool vdsm_util_parse_uri(const char *uri, char **host, char **port)
{
    *host = NULL;
    *port = NULL;

    if ((uri == NULL) || (strlen(uri) == 0)) {
        return FALSE;
    }

    if ((host == NULL) || (port == NULL)) {
        return FALSE;
    }

    char *delim = strrchr(uri, ':');

    /* uri can not start with ':' */
    if (uri == delim) {
        return FALSE;
    }

    /* no separator, use default port */
    if (delim == NULL) {
        *port = strdup(default_vdc_port);
        if (*port == NULL) {
            vdsm_report(LOG_ERR, "not enough memory for default port");
            return FALSE;
        }

        *host = strdup(uri);
        if (*host == NULL) {
            vdsm_report(LOG_ERR, "not enough memory for hostname with default port\n");
            free(*port);
            *port = NULL;
            return FALSE;
        }
        return TRUE;
    }

    /* separator available, but no port given: invalid uri */
    if ((delim + 1) == NULL) {
        return FALSE;
    }

    *port = strdup(delim + 1);
    if (*port == NULL) {
        vdsm_report(LOG_ERR, "not enough memory for port");
        return FALSE;
    }

    *host = strndup(uri, delim - uri);
    if (*host == NULL) {
        vdsm_report(LOG_ERR, "not enough memory for hostname\n");
        free(*port);
        *port = NULL;
        return FALSE;
    }

    return TRUE;
}

int openVdcSessionSocket(const char *vdc_url, const char *vdc_port)
{
    struct addrinfo hints;
    struct addrinfo *result, *rp;
    int sfd, s;

    /* Obtain address(es) matching host/port */
    memset(&hints, 0, sizeof(struct addrinfo));
    /* Allow IPv4 */
    hints.ai_family = AF_INET;
    /* tcp socket */
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = 0;

    s = getaddrinfo(vdc_url, vdc_port, &hints, &result);
    if (s != 0) {
        vdsm_report(LOG_ERR, "getaddrinfo error: %s\n", gai_strerror(s));
        return -1;
    }

    /* getaddrinfo() returns a list of address structures.
    Try each address until we successfully connect(2).
    If socket(2) (or connect(2)) fails, we (close the socket
    and) try the next address. */
    for (rp = result; rp != NULL; rp = rp->ai_next) {
        sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
        if (sfd == -1) {
            continue;
        }
        if (connect(sfd, rp->ai_addr, rp->ai_addrlen) != -1) {
            /* Success */
            break;
        }
        close(sfd);
    }

    if (rp == NULL) {
        /* No address succeeded */
        freeaddrinfo(result);
        vdsm_report(LOG_ERR, "Could not connect to vDC (url: %s, port: %s)\n", vdc_url, vdc_port);
        return -1;
    }

    /* No longer needed */
    freeaddrinfo(result);
    return sfd;
}

char *sock2ip(int fd)
{
    struct sockaddr fsin;
    socklen_t alen = (socklen_t)sizeof(fsin);
    char *ip;
    int r;

    r = getpeername(fd, (struct sockaddr *) &fsin, &alen);
    if (r == 0) {
        ip = inet_ntoa(((struct sockaddr_in *)(&fsin))->sin_addr);
    } else {
        vdsm_report(LOG_INFO, "getpeername() = %d, error = %s (%d)\n", r, strerror(errno), errno);
        ip = "<unknown>";
    }
    return ip;
}

void socketDetach(SocketClient* cl)
{
    char *c_ip;

    if (-1 == cl->fd)
        return;

    c_ip = sock2ip(cl->fd);
    shutdown(cl->fd, SHUT_RDWR);
    close(cl->fd);

    vdsm_report(LOG_INFO, "detaching %s (fd %d)\n",
           c_ip, cl->fd);

    cl->fd = -1;
    cl->active = 0;
}

#if defined HAVE_BACKTRACE && defined HAVE_BACKTRACE_SYMBOLS
void vdsm_util_print_backtrace()
{
    int i;
    void* b[512];
    int size = backtrace(b, 512);
    char **s = backtrace_symbols(b, size);
    for(i = 0; i < size; i++) {
        fprintf(stderr, "_STRACE_ %i %s\n", i, s[i]);
    }
    free(s);
    fprintf(stderr, "\n");
}
#endif

/**
 * void pointer copy function for unaligned memory (e.g. packed ds485_container_t)
 *
 * shield memcpy from knowing what the original type of arguments are.
 * If gcc 4.3 knows that src is a pointer to a struct, it will optimize memcpy using
 * __builtin_memcpy mechanism and WRONGLY ASSUME THAT src IS 32bit ALIGNED.
 * By wrapping memcpy in a not inlined function, compiler cannot deduce the origin of
 * src any more.
 */
void * __attribute__((noinline)) unaligned_memcpy(void *dst, const void *src, size_t n)
{
    return memcpy(dst, src, n);
}

static void printLfAndIndent(FILE *aOutFile, int aIndent)
{
    int i;
    fputc('\n', aOutFile);
    for (i = 0; i<aIndent; i++) fputc(' ', aOutFile);
}

// forward declaration
void protobufMessagePrintInternal(FILE *aOutFile, const ProtobufCMessage *aMessageP, int aIndent);

static void protobufFieldPrint(FILE *aOutFile, const ProtobufCFieldDescriptor *aFieldDescriptorP, const void *aData, size_t aIndex, int aIndent)
{
    switch (aFieldDescriptorP->type) {
    case PROTOBUF_C_TYPE_BOOL:
        fprintf(aOutFile, "(bool)%s", *((protobuf_c_boolean *)aData+aIndex) ? "true" : "false");
        break;
    case PROTOBUF_C_TYPE_INT32:
    case PROTOBUF_C_TYPE_SINT32:
    case PROTOBUF_C_TYPE_SFIXED32:
        fprintf(aOutFile, "(int32)%d", *((int32_t *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_INT64:
    case PROTOBUF_C_TYPE_SINT64:
    case PROTOBUF_C_TYPE_SFIXED64:
        fprintf(aOutFile, "(int64)%"PRId64, *((int64_t *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_UINT32:
    case PROTOBUF_C_TYPE_FIXED32:
        fprintf(aOutFile, "(uint32)%u", *((uint32_t *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_UINT64:
    case PROTOBUF_C_TYPE_FIXED64:
        fprintf(aOutFile, "(uint64)%"PRIu64, *((uint64_t *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_FLOAT:
        fprintf(aOutFile, "(float)%f", *((float *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_DOUBLE:
        fprintf(aOutFile, "(double)%f", *((double *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_ENUM:
        fprintf(aOutFile, "(enum)%d", *((int *)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_STRING:
        fprintf(aOutFile, "(string)\"%s\"", *((const char **)aData+aIndex));
        break;
    case PROTOBUF_C_TYPE_BYTES: {
        fprintf(aOutFile, "(bytes)");
        ProtobufCBinaryData *bd = ((ProtobufCBinaryData *)aData+aIndex);
        size_t n, i;
        n = (bd->len > 20) ? 20 : bd->len;
        for (i = 0; i < n; i++) {
            fprintf(aOutFile, "%02X", bd->data[i]);
        }
        if (n < bd->len) {
            fprintf(aOutFile, "... (%ld bytes)", bd->len);
        }
        break;
    }
    case PROTOBUF_C_TYPE_MESSAGE: {
        // submessage, pack into object value
        const ProtobufCMessage *subMessageP = *((const ProtobufCMessage **)aData+aIndex);
        protobufMessagePrintInternal(aOutFile, subMessageP, aIndent-2);
        break;
    }
    default:
        fprintf(aOutFile, "(unknown field type)");
        break;
    }
}

void protobufMessagePrintInternal(FILE *aOutFile, const ProtobufCMessage *aMessageP, int aIndent)
{
    const ProtobufCFieldDescriptor *fieldDescP;
    unsigned f;

    fprintf(aOutFile, "(%s) {", aMessageP->descriptor->name);
    aIndent += 2; // contents of message always indented
    // Fields
    fieldDescP = aMessageP->descriptor->fields;
    for (f = 0; f<aMessageP->descriptor->n_fields; f++) {
        const uint8_t *baseP = (const uint8_t *)(aMessageP);
        const uint8_t *fieldBaseP = baseP+fieldDescP->offset;
        if (fieldDescP->label==PROTOBUF_C_LABEL_REPEATED) {
            // repeated field, show all elements
            size_t arraySize = *((size_t *)(baseP+fieldDescP->quantifier_offset));
            size_t i;
            for (i = 0; i<arraySize; i++) {
                printLfAndIndent(aOutFile, aIndent);
                fprintf(aOutFile, "%s[%zu]: ", fieldDescP->name, i);
                protobufFieldPrint(aOutFile, fieldDescP, *((void **)fieldBaseP), i, aIndent+2); // every subsequent line of content indented
            }
        }
        else {
            // not repeated (array), but single field
            bool hasField = false;
            if (fieldDescP->label==PROTOBUF_C_LABEL_OPTIONAL) {
                if (fieldDescP->quantifier_offset) {
                    // scalar that needs quantifier
                    hasField = *((protobuf_c_boolean *)(baseP+fieldDescP->quantifier_offset));
                }
                else {
                    // value is a pointer, exists if not NULL
                    hasField = *((const void **)(baseP+fieldDescP->offset))!=NULL;
                }
            }
            else {
                // must be mandatory
                hasField = true;
            }
            // get value, if available
            if (hasField) {
                printLfAndIndent(aOutFile, aIndent);
                fprintf(aOutFile, "%s: ", fieldDescP->name);
                // get value
                protobufFieldPrint(aOutFile, fieldDescP, fieldBaseP, 0, aIndent+2); // every subsequent line of content indented
            }
        }
        fieldDescP++; // next field descriptor
    }
    aIndent -= 2; // end of message bracket unindented
    printLfAndIndent(aOutFile, aIndent);
    fprintf(aOutFile, "}");
}

void protobufMessagePrint(const ProtobufCMessage *aMessageP, int aIndent)
{
    if (LOG_DEBUG <= debugLevel) {
        pthread_mutex_lock(&reportMutex);
        // Start and end with a LF
        printLfAndIndent(stderr,aIndent);
        protobufMessagePrintInternal(stderr,aMessageP, aIndent);
        fprintf(stderr, "\n\n");
        pthread_mutex_unlock(&reportMutex);
    }
}

void convert_string_to_upper(char *data, size_t len)
{
    size_t ctr;
    for (ctr = 0; ctr < len; ++ctr) {
      data[ctr] = (toupper(data[ctr]));
    }
}

int dsuid_to_string_upper(const dsuid_t *dsuid, char *buf)
{
  int retVal = dsuid_to_string(dsuid, buf);
  convert_string_to_upper(buf,  DSUID_STR_LEN);
  return retVal;
}

