/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>
             Sergey Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <pthread.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <limits.h>
#include <math.h>
#include <uthash.h>
#include <utlist.h>

#ifdef HAVE_AVAHI
#include <avahi-client/client.h>
#include <avahi-client/lookup.h>

#include <avahi-common/simple-watch.h>
#include <avahi-common/malloc.h>
#include <avahi-common/error.h>
#include <avahi-common/domain.h>

#include "coordination.h"
#include "discovery.h"
#include "whitelist.h"

#define DEFAULT_AVAHI_SERVICE_NAME "digitalSTROM vdSM"

#endif

#include "dsm.h"
#include "selectB.h"
#include "vdcProcessor.h"
#include "dsmapi.h"
#include "registers.h"
#include "button.h"
#include "datamodel.h"
#include "sockutil.h"
#include "msgutil.h"
#include "properties.h"
#include "proputils.h"
#include "util.h"
#include "channelutil.h"
#include "vdc-api-proto.h"

#define VDC_SOCKET_READ_TIMEOUT     2       /* seconds */
#define VDC_SOCKET_WRITE_TIMEOUT    3       /* seconds */
#define PROTO_MESSAGE_ID_INIT       0
#define VDC_TIMEOUT_S               50      /* seconds */
#define MAX_DATA_SIZE               16384   /* accept msgs not larger than...*/

Q_DEFINE_THIS_MODULE(vdc-api)

typedef struct cached_request {
    struct cached_request* next;
    Vdcapi__Message *msg;
    request_cache_t *request_cache;
    time_t timestamp;
} cached_request_t;

typedef struct {
    char device_dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client;
    UT_hash_handle hh;
} vdc_device_lookup_hash_t;

typedef struct {
    char vdc_dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client;
    UT_hash_handle hh;
} vdc_lookup_hash_t;

/* globals */
static bool g_whitelist_changed = false;
static int g_shutdown_flag = 0;
static bool g_vdc_down = false;
static dsuid_t g_dsm_dsuid;
static uint32_t g_message_id = PROTO_MESSAGE_ID_INIT;

static vdc_client_t *g_vdc_client_list = NULL;

static pthread_rwlock_t g_vdc_client_list_mutex;

static pthread_mutex_t g_cached_request_list_mutex;
static cached_request_t *g_cached_request_list = NULL;

static vdc_device_lookup_hash_t *g_vdc_device_lookup_hash = NULL;
static pthread_mutex_t g_vdc_device_lookup_hash_mutex;

static vdc_lookup_hash_t * g_vdc_lookup_hash = NULL;
static pthread_mutex_t g_vdc_lookup_hash_mutex;

static pthread_t g_vdc_api_thread;

static uint16_t g_vdsm_listen_port = 0;
static VDSMRole g_vdsm_role;

//------------------------------------------------------------------------------
// declaration
static bool vdc_api_get_vdc_from_lookup_table_private(const char *dev_dsuid,
                                       vdc_client_t **vdc_client);

//------------------------------------------------------------------------------
int vdc_api_proto_LOCK()
{
    return pthread_rwlock_rdlock(&g_vdc_client_list_mutex);
}

int  vdc_api_proto_UNLOCK()
{
    return pthread_rwlock_unlock(&g_vdc_client_list_mutex);
}

//------------------------------------------------------------------------------
/**
 * @brief appendDsuidToVdcEntry internal use only. Append dsuid to vdc client list item
 * @param vdcClient pointer to list. Add element to it
 * @param dsuid the dsuid to be added
 */
static void appendDsuidToVdcEntry(vdc_dsuid_list_t *vdcClient, device_dsuid_t *dsuid)
{
    if (vdcClient && dsuid) {
        void *tmp = realloc(vdcClient->dsuid, (vdcClient->n_dsuid+1)*sizeof(char *));
        if (!tmp) {
            goto appendFail;
        }
        vdcClient->dsuid = tmp;

        char *newdsuid = malloc(DSUID_STR_LEN);
        if (!newdsuid) {
            goto appendFail;
        }
        strncpy(newdsuid, dsuid->dsuid, DSUID_STR_LEN); // including terminator

        vdcClient->dsuid[vdcClient->n_dsuid] = newdsuid;
        vdcClient->n_dsuid++;
        return;
    }
appendFail:
    vdsm_report(LOG_ERR, "not enough memory to create dsuid list entry for device %s\n", dsuid->dsuid);
}

/**
 * @brief findVdcInList find vdc_client in lvdcClientList. If not found, return NULL
 * @param vdcClientList
 * @param vdc_client
 * @return vdc_dsuid_list_t entry or null, if not found
 */
static vdc_dsuid_list_t *findVdcInList(vdc_dsuid_list_t *vdcClientList, vdc_client_t *vdc_client)
{
    vdc_dsuid_list_t *current_vdc, *tmp_vdc;
    LL_FOREACH_SAFE(vdcClientList, current_vdc, tmp_vdc) {
        if (current_vdc->vdc_client == vdc_client) {
            return current_vdc;
        }
    }
    return NULL;
}

void resortByVdcClient(device_dsuid_t *dsuid_list, vdc_dsuid_list_t **vdc_dsuid_list)
{
    device_dsuid_t *dsuid, *tmp_dsuid;
    vdc_client_t *vdc_client;

    // now find vdc client and vdc api side ID for each input dsuid
    LL_FOREACH_SAFE(dsuid_list, dsuid, tmp_dsuid) {
        if (vdc_api_get_vdc_from_lookup_table_private(dsuid->dsuid, &vdc_client)) {
            vdc_dsuid_list_t *tempItem = findVdcInList(*vdc_dsuid_list, vdc_client);
            if (!tempItem) {
                tempItem = (vdc_dsuid_list_t*) malloc(sizeof(vdc_dsuid_list_t));
                if (tempItem) {
                    tempItem->n_dsuid = 0; // none used yet
                    tempItem->dsuid = NULL;
                    tempItem->vdc_client = vdc_client;
                    LL_PREPEND(*vdc_dsuid_list, tempItem);
                } else {
                    vdsm_report(LOG_ERR, "not enough memory to create notification dsuid list for vdc host %s\n", vdc_client->dsuid);
                    // incomplete entry, free it
                    free(tempItem);
                    tempItem = NULL;
                }
            }
            if (tempItem) {
                appendDsuidToVdcEntry(tempItem, dsuid);
            }
        }
        LL_DELETE(dsuid_list, dsuid);
        free(dsuid);
    }
}

void free_vdc_item(vdc_dsuid_list_t *vdc_item)
{
    // the id strings
    size_t i;
    for (i=0; i < vdc_item->n_dsuid; ++i) {
        free(vdc_item->dsuid[i]);
    }
    // the list of ids
    free(vdc_item->dsuid);
    // the item itself
    free(vdc_item);
}

//------------------------------------------------------------------------------

static void vdc_api_free_device_lookup_table()
{
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);

    vdc_device_lookup_hash_t *vdc_device_lookup, *tmp;

    HASH_ITER(hh, g_vdc_device_lookup_hash, vdc_device_lookup, tmp) {
        HASH_DEL(g_vdc_device_lookup_hash, vdc_device_lookup);
        free(vdc_device_lookup);
    }

    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
}

static void vdc_api_add_device_to_vdc_lookup_table(vdc_client_t *vdc_client,
                                                   const char *dev_dsuid)
{
    vdc_device_lookup_hash_t *vdc_device_lookup;

    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    HASH_FIND_STR(g_vdc_device_lookup_hash, dev_dsuid, vdc_device_lookup);
    if (vdc_device_lookup == NULL) {
        vdc_device_lookup = malloc(sizeof(vdc_device_lookup_hash_t));
        if (vdc_device_lookup) {
            strncpy(vdc_device_lookup->device_dsuid, dev_dsuid, sizeof(vdc_device_lookup->device_dsuid));
            convert_string_to_upper(vdc_device_lookup->device_dsuid, sizeof(vdc_device_lookup->device_dsuid));
            HASH_ADD_STR(g_vdc_device_lookup_hash, device_dsuid,
                         vdc_device_lookup);
        } else {
            vdsm_report(LOG_ERR, "not enough memory to add device %s to lookup "
                            "hash\n", dev_dsuid);
            pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
            return;
        }
    }
    vdc_device_lookup->vdc_client = vdc_client;

    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
}

// lock mutex g_vdc_device_lookup_hash_mutex
static bool vdc_api_get_vdc_from_lookup_table_private(const char *dev_dsuid,
                                       vdc_client_t **vdc_client)
{
    vdc_device_lookup_hash_t *vdc_device_lookup;
    HASH_FIND_STR(g_vdc_device_lookup_hash, dev_dsuid, vdc_device_lookup);
    if (vdc_device_lookup != NULL) {
        *vdc_client = vdc_device_lookup->vdc_client;
        return TRUE;
    }
    vdsm_report(LOG_NOTICE, "vdc api get vdc from lookup table. device not "
                            "found %s\n", dev_dsuid);
    *vdc_client = NULL;
    return FALSE;
}

static void vdc_api_remove_devices_from_vdc_lookup_tables(vdc_client_t *vdc_client)
{
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);

    vdc_device_lookup_hash_t *vdc_device_lookup, *tmp;

    HASH_ITER(hh, g_vdc_device_lookup_hash, vdc_device_lookup, tmp) {
        if (vdc_device_lookup->vdc_client == vdc_client) {
          HASH_DEL(g_vdc_device_lookup_hash, vdc_device_lookup);
          vdsm_report(LOG_ERR, "remove device %s from vdc lookup. vdc host: %s\n",
              vdc_device_lookup->device_dsuid, vdc_device_lookup->vdc_client->dsuid);
          free(vdc_device_lookup);
        }
    }

    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);

    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);

    vdc_lookup_hash_t *vdc_lookup, *tmp2;

    HASH_ITER(hh, g_vdc_lookup_hash, vdc_lookup, tmp2) {
        if (vdc_lookup->vdc_client == vdc_client) {
            HASH_DEL(g_vdc_lookup_hash, vdc_lookup);
            vdsm_report(LOG_ERR, "remove vdc %s from vdc lookup. vdc host: %s\n",
                vdc_lookup->vdc_dsuid, vdc_lookup->vdc_client->dsuid);
            free(vdc_lookup);
        }
    }

    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
}

// vdc
static void vdc_api_free_vdc_lookup_table()
{
    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);

    vdc_lookup_hash_t *vdc_lookup, *tmp;

    HASH_ITER(hh, g_vdc_lookup_hash, vdc_lookup, tmp) {
        HASH_DEL(g_vdc_lookup_hash, vdc_lookup);
        free(vdc_lookup);
    }

    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
}

static void vdc_api_add_vdc_lookup_table(vdc_client_t *vdc_client,
                                         const char *vdc_dsuid)
{
    vdc_lookup_hash_t *vdc_lookup;

    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);
    HASH_FIND_STR(g_vdc_lookup_hash, vdc_dsuid, vdc_lookup);
    if (vdc_lookup == NULL) {
        vdc_lookup = malloc(sizeof(vdc_lookup_hash_t));
        if (vdc_lookup) {
            strncpy(vdc_lookup->vdc_dsuid, vdc_dsuid, sizeof(vdc_lookup->vdc_dsuid));
            convert_string_to_upper(vdc_lookup->vdc_dsuid, sizeof(vdc_lookup->vdc_dsuid));
            HASH_ADD_STR(g_vdc_lookup_hash, vdc_dsuid,
                         vdc_lookup);
        } else {
            vdsm_report(LOG_ERR, "not enough memory to add vdc %s to lookup "
                            "hash\n", vdc_dsuid);
            pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
            return;
        }
    }
    vdc_lookup->vdc_client = vdc_client;

    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
}

// lock mutex g_vdc_lookup_hash_mutex
static bool vdc_api_get_vdc_host_for_vdc_private(const char *dev_dsuid,
                                  vdc_client_t **vdc_client)
{
    vdc_lookup_hash_t *vdc_lookup;
    HASH_FIND_STR(g_vdc_lookup_hash, dev_dsuid, vdc_lookup);
    if (vdc_lookup != NULL) {
        *vdc_client = vdc_lookup->vdc_client;
        return TRUE;
    }

    vdsm_report(LOG_NOTICE, "vdc api get vdc host from vdc table. device not "
                      "found %s\n", dev_dsuid);
    *vdc_client = NULL;
    return FALSE;
}


int vdc_api_get_vdc_dsuid(dsuid_t *devices, int length)
{
    if (length < 1) return 0;

    int ctr = 0;

    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);
    dsuid_t dsuid;

    vdc_lookup_hash_t *vdc_lookup, *tmp;
    HASH_ITER(hh, g_vdc_lookup_hash, vdc_lookup, tmp) {
        dsuid_from_string(vdc_lookup->vdc_dsuid, &dsuid);
        memcpy(&devices[ctr], &dsuid, sizeof(dsuid_t));
        vdsm_report(LOG_DEBUG, "Devices: %s\n", vdc_lookup->vdc_dsuid);
        ++ctr;
        if (ctr >= length) break;
    }

    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
    return ctr;
}


void vdc_api_free_message(Vdcapi__Message *msg)
{
    if (!msg) {
        return;
    }

    switch (msg->type) {
        case VDCAPI__TYPE__VDSM_REQUEST_HELLO:
            if (msg->vdsm_request_hello) {
                if (msg->vdsm_request_hello->dsuid) {
                    free(msg->vdsm_request_hello->dsuid);
                }
                free(msg->vdsm_request_hello);
            }
            break;
        case VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY:
            if (msg->vdsm_request_get_property) {
                proputil_get_property_free(msg->vdsm_request_get_property);
            }
            break;
        case VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY:
            if (msg->vdsm_request_set_property) {
                proputil_set_property_free(msg->vdsm_request_set_property);
            }
            break;
        case VDCAPI__TYPE__VDSM_SEND_REMOVE:
            if (msg->vdsm_send_remove) {
                free(msg->vdsm_send_remove);
            }
            break;
        default:
            break;
    }

    free(msg);
}

static void vdc_api_set_vanish_vdc(vdc_client_t *vdc_client)
{
    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);

    vdc_lookup_hash_t *vdc_lookup, *tmp;

    HASH_ITER(hh, g_vdc_lookup_hash, vdc_lookup, tmp) {
        if (vdc_lookup->vdc_client == vdc_client) {
            DS485_EmitBusChange(vdc_lookup->vdc_dsuid, 1);
        }
    }
    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
}

static void vdc_api_reset_connection(vdc_client_t *vdc_client)
{
    vdc_api_set_vanish_vdc(vdc_client);

    pthread_mutex_lock(&vdc_client->mutex);
    if (vdc_client->fd != -1) {
        close(vdc_client->fd);
        vdc_client->fd = -1;
    }

    /* autodiscovered devices will be removed by avahi, but manually added
     * devices must not be removed (i.e. we must attempt to reconnect) but
     * we need to reset them completely, because there may be a totally new
     * vDC on the other end later on (i.e. with different dSUID, etc) */
    if (!vdc_client->autodiscovered) {
        memset(vdc_client->dsuid, 0, DSUID_STR_LEN);
    }
    pthread_mutex_unlock(&vdc_client->mutex);
}

static void vdc_api_send_generic_response(vdc_client_t *vdc_client,
                                          Vdcapi__ResultCode code, uint32_t id)
{
    if (id == 0) {
        vdsm_report(LOG_WARNING,
               "Not sending response to vdc: invalid message id\n");
        return;
    }

    Vdcapi__Message msg = VDCAPI__MESSAGE__INIT;
    Vdcapi__GenericResponse submsg = VDCAPI__GENERIC_RESPONSE__INIT;

    submsg.code = code;

    msg.message_id = id;
    msg.has_message_id = 1;
    msg.generic_response = &submsg;

    switch (code) {
    case VDCAPI__RESULT_CODE__ERR_OK:
        submsg.description = "OK";
        break;
    case VDCAPI__RESULT_CODE__ERR_MESSAGE_UNKNOWN:
        submsg.description = "Unknown message type";
        break;
    case VDCAPI__RESULT_CODE__ERR_INCOMPATIBLE_API:
        submsg.description = "Incompatible API version";
        break;
    case VDCAPI__RESULT_CODE__ERR_SERVICE_NOT_AVAILABLE:
        submsg.description = "Service not available";
        break;
    case VDCAPI__RESULT_CODE__ERR_INSUFFICIENT_STORAGE:
        submsg.description = "Insufficient storage";
        break;
    case VDCAPI__RESULT_CODE__ERR_FORBIDDEN:
        submsg.description = "Forbidden";
        break;
    case VDCAPI__RESULT_CODE__ERR_NOT_IMPLEMENTED:
        submsg.description = "Not implemented";
        break;
    case VDCAPI__RESULT_CODE__ERR_NO_CONTENT_FOR_ARRAY:
        submsg.description = "No content for array"; // do we need that?
        break;
    case VDCAPI__RESULT_CODE__ERR_INVALID_VALUE_TYPE:
        submsg.description = "Invalid or unexpected value type";
        break;
    case VDCAPI__RESULT_CODE__ERR_MISSING_SUBMESSAGE:
        submsg.description = "Missing protocol submessage";
        break;
    case VDCAPI__RESULT_CODE__ERR_MISSING_DATA:
        submsg.description = "Missing data / empty message";
        break;
    default:
        vdsm_report(LOG_INFO, "vdc api send error response. Response Error is not handled.\n");
    };
    vdc_api_send_message(vdc_client, &msg, NULL);
}

static void vdc_api_send_ok_response(vdc_client_t *vdc_client, uint32_t id)
{
    vdc_api_send_generic_response(vdc_client, VDCAPI__RESULT_CODE__ERR_OK, id);
}

static void vdc_api_process_error_response(vdc_client_t *vdc_client,
                                           Vdcapi__Message *cached_msg,
                                           Vdcapi__ResultCode res)
{
    switch (cached_msg->type)
    {
        case VDCAPI__TYPE__VDC_RESPONSE_HELLO:
            vdsm_report(LOG_NOTICE, "vDC host %s refused us :(\n", vdc_client->dsuid);
            /* someone is probably already connected to this vdc */
            if (res == VDCAPI__RESULT_CODE__ERR_SERVICE_NOT_AVAILABLE) {
                pthread_mutex_lock(&vdc_client->mutex);
                vdc_api_reset_connection(vdc_client);
                pthread_mutex_unlock(&vdc_client->mutex);
            }
        default:
            break;
    }
}

static int vdc_api_cache_request(Vdcapi__Message *msg, request_cache_t *rc)
{
    vdsm_report(LOG_DEBUG, "caching request with id %u\n", msg->message_id);
    cached_request_t *proto_request = malloc(sizeof(cached_request_t));
    if (proto_request == NULL) {
        vdsm_report(LOG_ERR, "coult not allocate memory for request cache\n");
        return VDC_API_ERR_OUT_OF_MEMORY;
    }

    proto_request->msg = msg;
    proto_request->request_cache = rc;
    proto_request->timestamp = time(NULL);
    pthread_mutex_lock(&g_cached_request_list_mutex);
    LL_PREPEND(g_cached_request_list, proto_request);
    pthread_mutex_unlock(&g_cached_request_list_mutex);

    return VDC_API_OK;
}

static bool vdc_api_peek_request_cache(uint32_t message_id,
                                       request_cache_t **rc)
{
    /* get rc for cached message but leave content in g_cached_request_list */
    cached_request_t *proto_request = NULL;
    cached_request_t *tmp = NULL;

    bool found = FALSE;
    pthread_mutex_lock(&g_cached_request_list_mutex);
    LL_FOREACH_SAFE(g_cached_request_list, proto_request, tmp) {
        if (proto_request->msg &&
            (proto_request->msg->message_id == message_id))
        {
            // found request with matching id
            found = TRUE;

            if (rc != NULL) {
                *rc = proto_request->request_cache;
            }
            break;
        }
    }
    pthread_mutex_unlock(&g_cached_request_list_mutex);
    return found;
}

static bool vdc_api_get_cached_request(uint32_t message_id,
                                       Vdcapi__Message **msg,
                                       request_cache_t **rc)
{
    cached_request_t *proto_request = NULL;
    cached_request_t *tmp = NULL;

    bool found = FALSE;

    pthread_mutex_lock(&g_cached_request_list_mutex);
    LL_FOREACH_SAFE(g_cached_request_list, proto_request, tmp) {
        if (proto_request->msg &&
            (proto_request->msg->message_id == message_id))
        {
            // found request with matching id
            found = TRUE;
            if (msg != NULL) {
                *msg = proto_request->msg;
            } else {
                vdc_api_free_message(proto_request->msg);
            }

            if (rc != NULL) {
                *rc = proto_request->request_cache;
            } else {
                if (proto_request->request_cache) {
                    free(proto_request->request_cache);
                }
            }
            LL_DELETE(g_cached_request_list, proto_request);
            free(proto_request);
            break;
        }
    }
    pthread_mutex_unlock(&g_cached_request_list_mutex);
    return found;
}

static void vdc_api_free_message_cache(bool timeout)
{
    cached_request_t *proto_request = NULL;
    cached_request_t *tmp = NULL;

    pthread_mutex_lock(&g_cached_request_list_mutex);
    LL_FOREACH_SAFE(g_cached_request_list, proto_request, tmp) {
        if ((!timeout) || (timeout &&
            ((uint16_t)difftime(time(NULL), proto_request->timestamp)
             >= VDC_TIMEOUT_S)))
        {
            vdc_api_free_message(proto_request->msg);
            if (proto_request->request_cache) {
                free(proto_request->request_cache);
            }
            LL_DELETE(g_cached_request_list, proto_request);
            free(proto_request);
        }
    }
    pthread_mutex_unlock(&g_cached_request_list_mutex);
}

static int vdc_api_msg_is_request(Vdcapi__Type t)
{
    switch (t) {
        case VDCAPI__TYPE__VDSM_REQUEST_HELLO:
        case VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY:
        case VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY:
        case VDCAPI__TYPE__VDSM_SEND_REMOVE:
            return 1;
        default:
            return 0;
    };
}

static void vdc_api_process_announce_device(vdc_client_t *vdc_client,
                                     Vdcapi__Message *msg)
{
    if (msg->vdc_send_announce_device == NULL) {
        vdsm_report(LOG_ERR, "Received announce packet, but message is missing\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_SUBMESSAGE, msg->message_id);
        return;
    }

    if (!msg->vdc_send_announce_device->dsuid) {
        vdsm_report(LOG_ERR, "Announce message is missing dSUID parameter\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_DATA, msg->message_id);
        return;
    }

    if (!msg->vdc_send_announce_device->vdc_dsuid) {
        vdsm_report(LOG_ERR, "Announce message is missing vdc dSUID parameter\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_DATA, msg->message_id);
        return;
    }

    char dsuid_upper[DSUID_STR_LEN];
    char vdc_dsuid_upper[DSUID_STR_LEN];
    strncpy(dsuid_upper, msg->vdc_send_announce_device->dsuid, DSUID_STR_LEN);
    strncpy(vdc_dsuid_upper, msg->vdc_send_announce_device->vdc_dsuid, DSUID_STR_LEN);
    convert_string_to_upper(dsuid_upper,  DSUID_STR_LEN);
    convert_string_to_upper(vdc_dsuid_upper,  DSUID_STR_LEN);


    vdsm_report(LOG_NOTICE, "Announce msg id %d parameter dSUID read: %s, vdc dSUID"
                      " %s\n", msg->message_id, dsuid_upper, vdc_dsuid_upper);

    vdc_api_add_device_to_vdc_lookup_table(vdc_client, dsuid_upper);
    SelectB_AnnounceDevice(dsuid_upper, vdc_dsuid_upper, msg->message_id);
}

void vdc_api_process_announce_vdc(vdc_client_t *vdc_client,
                                      Vdcapi__Message *msg)
{
    if (msg->vdc_send_announce_vdc == NULL) {
        vdsm_report(LOG_ERR, "Received announce  vdc packet, but message is missing\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_SUBMESSAGE, msg->message_id);
        return;
    }

    if (!msg->vdc_send_announce_vdc->dsuid) {
        vdsm_report(LOG_ERR, "Announce vdc message is missing dSUID parameter\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_DATA, msg->message_id);
        return;
    }

    char dsuid_upper[DSUID_STR_LEN];
    strncpy(dsuid_upper, msg->vdc_send_announce_vdc->dsuid, DSUID_STR_LEN);
    convert_string_to_upper(dsuid_upper,  DSUID_STR_LEN);

    vdc_api_add_vdc_lookup_table(vdc_client, dsuid_upper);
    Datamodel_add_vdc(dsuid_upper);

    VDC_AnnounceVDC(dsuid_upper, msg->message_id);
}

static void vdc_api_process_pong(vdc_client_t *vdc_client,
                                  Vdcapi__Message *msg)
{
    if (msg->vdc_send_pong == NULL) {
        vdsm_report(LOG_ERR, "Received pong packet, but message is missing\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_SUBMESSAGE,
                msg->message_id);
        return;

    }

    if (!msg->vdc_send_pong->dsuid) {
        vdsm_report(LOG_ERR, "Pong message is missing dSUID parameter\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_DATA,
                msg->message_id);
        return;
    }

    char dsuid_upper[DSUID_STR_LEN];
    strncpy(dsuid_upper, msg->vdc_send_pong->dsuid, DSUID_STR_LEN);
    convert_string_to_upper(dsuid_upper,  DSUID_STR_LEN);

    uint16_t device_id;
    vdsm_report(LOG_DEBUG, "Pong notification received from %s\n", dsuid_upper);
    device_id = Datamodel_GetDeviceIdFromDsuid(dsuid_upper);
    if (device_id != 1023 ) {
        PingResponseEvt * e = Q_NEW(PingResponseEvt,
                SELECT_B_PING_RESPONSE_SIG);
        e->deviceId = device_id;
        QActive_postFIFO(AO_SelectB, (QEvent *)e);
        vdsm_report(LOG_DEBUG, "Sent ping response for device %d\n", device_id);
    }
}

static void vdc_api_process_vanish(vdc_client_t *vdc_client,
                                   Vdcapi__Message *msg)
{
    vdsm_report(LOG_DEBUG, "Vanish message received\n");
    if (msg->vdc_send_vanish == NULL) {
        vdsm_report(LOG_ERR, "Received vanish packet, but message is missing\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_SUBMESSAGE,
                msg->message_id);
        return;
    }

    if (!msg->vdc_send_vanish->dsuid) {
        vdsm_report(LOG_ERR, "Vanish message is missing dSUID parameter\n");
        vdc_api_send_generic_response(vdc_client,
                VDCAPI__RESULT_CODE__ERR_MISSING_DATA,
                msg->message_id);
        return;
    }

    char dsuid_upper[DSUID_STR_LEN];
    strncpy(dsuid_upper, msg->vdc_send_vanish->dsuid, DSUID_STR_LEN);
    convert_string_to_upper(dsuid_upper,  DSUID_STR_LEN);

    uint16_t device_id;
    vdsm_report(LOG_NOTICE, "Vanish notification received from %s\n", dsuid_upper);
    device_id = Datamodel_GetDeviceIdFromDsuid(dsuid_upper);

    // send device inactive to datamodel
    DatamodelReqPingDeviceStateEvt * reqEvt =
        Q_NEW(DatamodelReqPingDeviceStateEvt,
                DATAMODEL_REQ_PING_DEVICE_STATE_SIG);
    reqEvt->basic.requester = NULL;
    reqEvt->deviceId = device_id;
    reqEvt->active = FALSE;
    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
}

static void vdc_api_process_generic_response(vdc_client_t *vdc_client,
                                             Vdcapi__Message *msg)
{
    if (msg->generic_response == NULL) {
        vdsm_report(LOG_ERR, "received generic response packet, but "
                        "message is missing\n");
        /* this will free cached data */
        vdc_api_get_cached_request(msg->message_id, NULL, NULL);
        return;
    }

    const char* error = msg->generic_response->description;
    Vdcapi__ResultCode res = msg->generic_response->code;
    if (res == VDCAPI__RESULT_CODE__ERR_OK) {
        vdsm_report(LOG_DEBUG, "success response with id %d received\n",
               msg->message_id);

        /* this will free cached data */
        vdc_api_get_cached_request(msg->message_id, NULL, NULL);
    } else {
        vdsm_report(LOG_INFO, "error with id %u received: %s/%d\n",
               msg->message_id, error, res);
        Vdcapi__Message *cached_msg;
        if (vdc_api_get_cached_request(msg->message_id, &cached_msg, NULL)) {
            vdc_api_process_error_response(vdc_client, cached_msg, res);
            vdc_api_free_message(cached_msg);
        }
    }
}

static void vdc_api_process_hello(vdc_client_t *vdc_client,
                                  Vdcapi__Message *msg)
{
    Vdcapi__Message *cached_msg;
    if (!vdc_api_get_cached_request(msg->message_id, &cached_msg, NULL)) {
        vdsm_report(LOG_ERR, "received hello response, but request was "
                        "not found in cache\n");
        return;
    }

    if (msg->vdc_response_hello == NULL) {
        vdsm_report(LOG_ERR, "received hello response packet, but message "
                        "is missing\n");
        vdc_api_free_message(cached_msg);
        return;
    }

    if (msg->vdc_response_hello->dsuid == NULL) {
        vdsm_report(LOG_ERR, "received hello response packet, but dsuid "
                        "is missing\n");
        vdc_api_free_message(cached_msg);
        return;
    }

    pthread_mutex_lock(&vdc_client->mutex);
    if (vdc_client->autodiscovered) {
        /* The dsuid was provided by avahi. It should match the one in the message. */
        char msg_dsuid[DSUID_STR_LEN];
        strncpy(msg_dsuid, msg->vdc_response_hello->dsuid, DSUID_STR_LEN - 1);
        msg_dsuid[DSUID_STR_LEN-1] = '\0';
        convert_string_to_upper(msg_dsuid, DSUID_STR_LEN);
        if (strncmp(vdc_client->dsuid, msg_dsuid, DSUID_STR_LEN - 1) != 0) {
            vdsm_report(LOG_ERR, "vDC host dsuid mismatch:\n");
            vdsm_report(LOG_ERR, "   avahi dsuid: %s\n", vdc_client->dsuid);
            vdsm_report(LOG_ERR, "   hello dsuid: %s\n", msg_dsuid);
        }
    } else {
        strncpy(vdc_client->dsuid, msg->vdc_response_hello->dsuid, DSUID_STR_LEN-1);
        vdc_client->dsuid[DSUID_STR_LEN-1] = '\0';
        convert_string_to_upper(vdc_client->dsuid, DSUID_STR_LEN);
    }
    pthread_mutex_unlock(&vdc_client->mutex);

    vdsm_report(LOG_DEBUG, "vDC host %s is connected\n", vdc_client->dsuid);

    vdc_api_free_message(cached_msg);
}

static void vdc_api_process_protobuf_message_response(Vdcapi__Message *msg,
                                                      Vdcapi__Message *cached_msg,
                                                      request_cache_t *rc)
{
    size_t msg_len = vdcapi__message__get_packed_size(msg);
    uint8_t *msg_buf = malloc(msg_len * sizeof(uint8_t));
    if (!msg_buf) {
        vdsm_report(LOG_ERR, "could not allocate %zu bytes for protobuf message\n",
               msg_len);
        vdcapi__message__free_unpacked(msg, NULL);
        vdc_api_free_message(cached_msg);
        free(rc);
        return;
    }
    msg_len = vdcapi__message__pack(msg, msg_buf);
    SelectB_UserProtobufMessageResponse(msg_len, msg_buf);
    vdcapi__message__free_unpacked(msg, NULL);
    vdc_api_free_message(cached_msg);
    free(msg_buf);
    free(rc);
}

static void vdc_api_process_get_property(vdc_client_t *vdc_client,
                                         Vdcapi__Message *msg)
{
    if (msg->vdc_response_get_property == NULL) {
        vdsm_report(LOG_ERR, "received getProperty response packet, but "
                        "message is missing\n");
        /* free cached request */
        vdc_api_get_cached_request(msg->message_id, NULL, NULL);
        return;
    }

    request_cache_t *rc;
    Vdcapi__Message *cached_msg;

    if (!vdc_api_get_cached_request(msg->message_id, &cached_msg, &rc)) {
        vdsm_report(LOG_ERR, "received getProperty response, but request was "
                        "not found in cache\n");
        return;
    }

    if (!cached_msg) {
        vdsm_report(LOG_ERR, "received getProperty response, but request messages "
                        "was not found in cache\n");
        if (rc) {
            free(rc);
        }
        return;
    }

    /* Request cache may be reused together with multiple request, so we
     * do not know if we are allowed to free it here. The functions that
     * process the properties will take care of deallocating it. */
    if (rc) {
        prop_process_properties(vdc_client, msg->vdc_response_get_property,
                                cached_msg, rc);
    }

    vdc_api_free_message(cached_msg);
    return;

    // TODO: process other properties / user properties
}

static void vdc_api_send_ping(vdc_client_t *vdc_client)
{
    Vdcapi__Message msg = VDCAPI__MESSAGE__INIT;
    Vdcapi__VdsmSendPing ping_submsg = VDCAPI__VDSM__SEND_PING__INIT;

    ping_submsg.dsuid = vdc_client->dsuid;
    msg.type = VDCAPI__TYPE__VDSM_SEND_PING;
    msg.vdsm_send_ping = &ping_submsg;

    // static message, not stored for response tracking
    vdc_api_send_message(vdc_client, &msg, NULL);
}

static void vdc_api_send_hello(vdc_client_t *vdc_client)
{
    Vdcapi__Message *msg;
    Vdcapi__VdsmRequestHello *hello_submsg;

    msg = malloc(sizeof(Vdcapi__Message));
    if (msg == NULL) {
        vdsm_report(LOG_ERR, "vdc api send hello: failed to allocate memory.\n");
        return;
    }

    hello_submsg = malloc(sizeof(Vdcapi__VdsmRequestHello));
    if (hello_submsg == NULL) {
        vdsm_report(LOG_ERR, "vdc api send hello: failed to allocate memory.\n");
        free(msg);
        return;
    }

    vdcapi__message__init(msg);
    vdcapi__vdsm__request_hello__init(hello_submsg);

    msg->type = VDCAPI__TYPE__VDSM_REQUEST_HELLO;

    char dsm_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(&g_dsm_dsuid, dsm_dsuid_str);
    hello_submsg->dsuid = strdup(dsm_dsuid_str);
    if (hello_submsg->dsuid == NULL) {
        vdsm_report(LOG_ERR, "vdc api send hello: failed to allocate memory.\n");
        free(hello_submsg);
        free(msg);
        return;
    }

    hello_submsg->api_version = 2;
    hello_submsg->has_api_version = 1;
    msg->vdsm_request_hello = hello_submsg;

    // message must be allocated because this request is stored
    // it will be freed once it is no longer needed
    vdc_api_send_message(vdc_client, msg, NULL);
}


static bool vdc_api_connect_client(vdc_client_t *vdc_client)
{
    bool ret = FALSE;
    pthread_mutex_lock(&vdc_client->mutex);
    if ((uint16_t)difftime(time(NULL), vdc_client->last_activity) >=
                                                    VDC_TIMEOUT_S) {
        int sfd;

        /* if client was conneted somehow, disconnect it first */
        if (vdc_client->fd != -1) {
            close(vdc_client->fd);
            vdc_client->fd = -1;
        }

        /* save this attempt as last active time */
        time(&vdc_client->last_activity);
        /* open vdc session socket */
        sfd = openVdcSessionSocket(vdc_client->uri, vdc_client->port);
        if (sfd != -1) {
            vdc_client->fd = sfd;
            vdsm_report(LOG_NOTICE, "start vDC connection to %s:%s\n",
                   vdc_client->uri, vdc_client->port);
            /* send hello to vDC */
            vdc_api_send_hello(vdc_client);
            ret = TRUE;
        } else {
            vdsm_report(LOG_NOTICE, "could not connect to vDC on %s at port %s\n",
                   vdc_client->uri, vdc_client->port);
        }
    }
    pthread_mutex_unlock(&vdc_client->mutex);
    return ret;
}

static void vdc_api_check_client_activity(vdc_client_t *vdc_client)
{
    pthread_mutex_lock(&vdc_client->mutex);
    /* send ping if vdc was inactive */
    if ((uint16_t)difftime(time(NULL), vdc_client->last_activity) >=
                                                        VDC_TIMEOUT_S) {
        vdc_api_send_ping(vdc_client);
        time(&vdc_client->last_activity);
    }
    pthread_mutex_unlock(&vdc_client->mutex);
}

static void vdc_api_process_message(vdc_client_t *vdc_client,
                             const unsigned char *data, const uint16_t len)
{
    Vdcapi__Message *msg = vdcapi__message__unpack(NULL, len, data);
    if (!msg)
    {
        vdsm_report(LOG_ERR, "failed to unpack incoming message\n");
        return;
    }
    vdsm_report(LOG_DEBUG, "Receive message \n");
    protobufMessagePrint(&msg->base, 0);

    /* Intercept dsm protobuf messages */
    request_cache_t *rc = NULL;
    Vdcapi__Message *cached_msg;
    if (vdc_api_peek_request_cache(msg->message_id, &rc)) {
        if (rc && rc->context == REQUEST_CONTEXT_GET_PROTOBUF_DATA) {
            vdc_api_get_cached_request(msg->message_id, &cached_msg, &rc);
            vdc_api_process_protobuf_message_response(msg, cached_msg, rc);
            return;
        }
    }

    switch (msg->type)
    {

        case VDCAPI__TYPE__VDC_SEND_ANNOUNCE_VDC:
            vdc_api_process_announce_vdc(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_SEND_ANNOUNCE_DEVICE:
            vdc_api_process_announce_device(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_SEND_PONG:
            vdc_api_process_pong(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_SEND_VANISH:
            vdc_api_process_vanish(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_SEND_IDENTIFY:
            prop_process_identify(msg);
            break;

        case VDCAPI__TYPE__VDC_SEND_PUSH_PROPERTY:
            prop_process_push_property(msg); /* handled by properties module */
            break;

        case VDCAPI__TYPE__GENERIC_RESPONSE:
            vdc_api_process_generic_response(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_RESPONSE_HELLO:
            vdc_api_process_hello(vdc_client, msg);
            break;

        case VDCAPI__TYPE__VDC_RESPONSE_GET_PROPERTY:
            vdc_api_process_get_property(vdc_client, msg);
            break;

        default:
            vdsm_report(LOG_ERR, "received call type from vDC is not valid: %d\n",
                   msg->type);
            /* try to clear that message from cache */
            if (msg->message_id != PROTO_MESSAGE_ID_INIT) {
                vdc_api_get_cached_request(msg->message_id, NULL, NULL);
            }
            break;
    }
    vdcapi__message__free_unpacked(msg, NULL);
}

static void vdc_api_iterate_socket()
{
    fd_set rfds;
    int max_fd;
    int retcode;
    vdc_client_t *vdc_client;

    max_fd = 0;
    FD_ZERO(&rfds);

    pthread_rwlock_rdlock(&g_vdc_client_list_mutex);

    LL_FOREACH(g_vdc_client_list, vdc_client) {
        pthread_mutex_lock(&vdc_client->mutex);

        if (vdc_client->fd != -1) {
            FD_SET(vdc_client->fd, &rfds);
            max_fd = max(vdc_client->fd, max_fd);
            vdc_api_check_client_activity(vdc_client);
        } else {
            if (vdc_api_connect_client(vdc_client)) {
                FD_SET(vdc_client->fd, &rfds);
                max_fd = max(vdc_client->fd, max_fd);
            }
        }

        pthread_mutex_unlock(&vdc_client->mutex);
    } /* LL_FOREACH */

    struct timeval tv;
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    retcode = select(max_fd + 1, &rfds, NULL, NULL, &tv);

    /* error (-1) or nothing set (0) */
    if (retcode <= 0) {
        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
        return;
    }

    LL_FOREACH(g_vdc_client_list, vdc_client) {
        pthread_mutex_lock(&vdc_client->mutex);

        /* client is not connected */
        if (vdc_client->fd == -1) {
            pthread_mutex_unlock(&vdc_client->mutex);
            continue;
        }

        if (!FD_ISSET(vdc_client->fd, &rfds)) {
            pthread_mutex_unlock(&vdc_client->mutex);
            continue;
        }

        uint16_t size;
        size_t len;
        retcode = sockread(vdc_client->fd, (unsigned char *)(&size),
                       sizeof(uint16_t), VDC_SOCKET_READ_TIMEOUT, &len);

        /* if things went south, close the socket, the vdc will be reconnected
           in the next iteration */
        if ((retcode != socket_ok) || (len != sizeof(uint16_t)) ||
           (ntohs(size) > MAX_DATA_SIZE))
        {
            vdc_api_reset_connection(vdc_client);
            vdsm_report(LOG_ERR, "could not read incoming message length or message "
                            "exceeds allowed size, resetting connection "
                            "for %s:%s (rc %d, length %d, size %d).\n",
                            vdc_client->uri, vdc_client->port, retcode, len, ntohs(size));
            pthread_mutex_unlock(&vdc_client->mutex);
            continue;
        }

        size = ntohs(size);
        if (size > 0)
        {
            unsigned char *data = malloc(size);
            if (!data)
            {
                vdc_api_reset_connection(vdc_client);
                vdsm_report(LOG_ERR, "could not allocate memory for incoming "
                                "message, resetting connection for %s:%s\n",
                                vdc_client->uri, vdc_client->port);
                pthread_mutex_unlock(&vdc_client->mutex);
                continue;
            }

            retcode = sockread(vdc_client->fd, data, size,
                               VDC_SOCKET_READ_TIMEOUT, &len);
            if ((retcode != socket_ok) || ((uint16_t)len != size))
            {
                vdc_api_reset_connection(vdc_client);
                free(data);

                vdsm_report(LOG_ERR, "could not read incoming message, resetting "
                       "connection for %s:%s (rc %d, length %d, size %d)\n",
                       vdc_client->uri, vdc_client->port, retcode, len, ntohs(size));
                pthread_mutex_unlock(&vdc_client->mutex);
                continue;
            }
            vdc_api_process_message(vdc_client, data, size);
            /* update activity timestamp */
            time(&vdc_client->last_activity);

            free(data);
        }

        pthread_mutex_unlock(&vdc_client->mutex);
    } /* LL_FOREACH */
    pthread_rwlock_unlock(&g_vdc_client_list_mutex);
}

static void vdc_api_free_clients()
{
    vdc_client_t *vdc_client = NULL;
    vdc_client_t *tmp = NULL;

    pthread_rwlock_wrlock(&g_vdc_client_list_mutex);

    LL_FOREACH_SAFE(g_vdc_client_list, vdc_client, tmp) {
        pthread_mutex_lock(&vdc_client->mutex);

        LL_DELETE(g_vdc_client_list, vdc_client);
        // TODO: implement and send "bye"
        if (vdc_client->fd != -1) {
            close(vdc_client->fd);
        }

        if (vdc_client->uri) {
            free(vdc_client->uri);
        }

        if (vdc_client->port) {
            free(vdc_client->port);
        }

        if (vdc_client->avahi_service_name) {
            free(vdc_client->avahi_service_name);
        }

        pthread_mutex_unlock(&vdc_client->mutex);
        pthread_mutex_destroy(&vdc_client->mutex);
        free(vdc_client);
    } /* LL_FOREACH_SAFE */

    pthread_rwlock_unlock(&g_vdc_client_list_mutex);
}

#ifdef HAVE_AVAHI
static int vdc_api_avahi_init(vdsm_avahi_handle_t *avahi, char *name)
{
    vdsm_coordination_set_collection_initialized_callback(
        &vdsm_collection_initialized_callback);

    avahi->simple_poll = NULL;
    avahi->client = NULL;
    avahi->browser = NULL;
    avahi->coordination_browser = NULL;
    avahi->initialized = FALSE;
    avahi->vdsm_port = g_vdsm_listen_port;
    avahi->vdsm_role = g_vdsm_role;

    char dsm_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(&g_dsm_dsuid, dsm_dsuid_str);
    avahi->vdsm_dsuid = strdup(dsm_dsuid_str);
    if (!avahi->vdsm_dsuid) {
        vdsm_report(LOG_ERR, "could not allocate memory for avahi service dSUID\n");
        return VDC_API_ERR_SYSTEM;
    }

    avahi->name = strdup(name);
    if (!avahi->name) {
        vdsm_report(LOG_ERR, "could not allocate memory for avahi service name\n");
        free(avahi->vdsm_dsuid);
        return VDC_API_ERR_SYSTEM;
    }

    int error;

    avahi->simple_poll = avahi_simple_poll_new();
    if (!avahi->simple_poll) {
        vdsm_report(LOG_ERR, "could not allocate avahi_poll object\n");
        free(avahi->name);
        free(avahi->vdsm_dsuid);
        return VDC_API_ERR_SYSTEM;
    }

    avahi->client = avahi_client_new(avahi_simple_poll_get(avahi->simple_poll),
                                    0, vdsm_avahi_client_callback,
                                    avahi, &error);
    if (!avahi->client) {
        vdsm_report(LOG_ERR, "could not create avahi client: %s\n",
               avahi_strerror(error));
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
        free(avahi->name);
        free(avahi->vdsm_dsuid);
        return VDC_API_ERR_SYSTEM;
    }

    if (avahi->vdsm_role.collecting) {
        avahi->browser = avahi_service_browser_new(avahi->client, AVAHI_IF_UNSPEC,
                                                   AVAHI_PROTO_UNSPEC,
                                                   "_ds-vdc._tcp",
                                                   NULL, 0,
                                                   vdsm_avahi_browse_callback,
                                                   avahi);
        if (!avahi->browser) {
            vdsm_report(LOG_ERR, "could not create avahi browser: %s\n",
                        avahi_strerror(avahi_client_errno(avahi->client)));
            avahi_client_free(avahi->client);
            avahi->client = NULL;
            avahi_simple_poll_free(avahi->simple_poll);
            avahi->simple_poll = NULL;
            free(avahi->name);
            free(avahi->vdsm_dsuid);
            return VDC_API_ERR_SYSTEM;
        }

        avahi->coordination_browser = avahi_service_browser_new(
            avahi->client,
            AVAHI_IF_UNSPEC,
            AVAHI_PROTO_UNSPEC,
            "_ds-vdsm._tcp",
            NULL,
            0,
            vdsm_avahi_browse_callback,
            avahi);

        if (!avahi->coordination_browser) {
            vdsm_report(LOG_ERR, "could not create avahi coordination_browser: %s\n",
                        avahi_strerror(avahi_client_errno(avahi->client)));
            avahi_service_browser_free(avahi->browser);
            avahi->browser = NULL;
            avahi_client_free(avahi->client);
            avahi->client = NULL;
            avahi_simple_poll_free(avahi->simple_poll);
            avahi->simple_poll = NULL;
            free(avahi->name);
            free(avahi->vdsm_dsuid);
            return VDC_API_ERR_SYSTEM;
        }
    }

    avahi->initialized = TRUE;

    return VDC_API_OK;
}

static void vdc_api_avahi_shutdown(vdsm_avahi_handle_t *avahi)
{
    if (avahi->group) {
        avahi_entry_group_free(avahi->group);
        avahi->group = NULL;
    }

    if (avahi->browser) {
        avahi_service_browser_free(avahi->browser);
        avahi->browser = NULL;
    }

    if (avahi->coordination_browser) {
        avahi_service_browser_free(avahi->coordination_browser);
        avahi->coordination_browser = NULL;
    }

    if (avahi->client) {
        avahi_client_free(avahi->client);
        avahi->client = NULL;
    }

    if (avahi->simple_poll) {
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
    }

    if (avahi->vdsm_dsuid) {
        free(avahi->vdsm_dsuid);
        avahi->vdsm_dsuid = NULL;
    }

    if (avahi->name) {
        free(avahi->name);
        avahi->name = NULL;
    }
}
#endif

static void *vdc_api_service(void *arg)
{
#ifdef HAVE_AVAHI
    vdsm_avahi_handle_t avahi;

    if (vdc_api_avahi_init(&avahi, DEFAULT_AVAHI_SERVICE_NAME) != VDC_API_OK) {
        vdsm_report(LOG_ERR, "automatic device discovery disabled - avahi "
                        "initialization failed\n");
    }
#endif

    while (!g_shutdown_flag) {
        vdc_api_iterate_socket();

#ifdef HAVE_AVAHI
        if (avahi.initialized) {
            if (avahi_simple_poll_iterate(avahi.simple_poll, 0) != 0) {
                vdsm_report(LOG_ERR, "avahi error, reinitializing...\n");
                char *temp_name = strdup(avahi.name);
                vdc_api_avahi_shutdown(&avahi);
                vdc_api_avahi_init(&avahi, (temp_name != NULL) ?
                                   temp_name : DEFAULT_AVAHI_SERVICE_NAME);
                if (temp_name) {
                    free(temp_name);
                }
            }
            if (g_whitelist_changed) {
                g_whitelist_changed = false;
                vdsm_whitelist_changed_callback();
            }
        }
#endif

        vdc_api_free_message_cache(TRUE); /* free too old cached messages */
    }

#ifdef HAVE_AVAHI
    vdc_api_avahi_shutdown(&avahi);
#endif
    vdc_api_free_clients();
    vdc_api_free_device_lookup_table();
    vdc_api_free_vdc_lookup_table();
    vdc_api_free_message_cache(FALSE); /* free all cached messages */
    // free cached messages
    g_vdc_down = true;
    return arg;
}

static int vdc_api_allocate_client(vdc_client_t **vdc_client)
{
    *vdc_client = NULL;

    pthread_mutexattr_t attr;
    if (pthread_mutexattr_init(&attr) != 0) {
        vdsm_report(LOG_ERR, "failed to init mutex attribute: %s\n",
               strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0) {
        pthread_mutexattr_destroy(&attr);
        vdsm_report(LOG_ERR, "failed to set mutex type: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    vdc_client_t *client = malloc(sizeof(vdc_client_t));
    if (client == NULL) {
        vdsm_report(LOG_ERR, "failed to allocate memory for vDC client\n");
        pthread_mutexattr_destroy(&attr);
        return VDC_API_ERR_OUT_OF_MEMORY;
    }

    memset(client, 0, sizeof(vdc_client_t));
    client->fd = -1;
    if (pthread_mutex_init(&client->mutex, &attr) != 0) {
        vdsm_report(LOG_ERR, "failed to init vdc client mutex: %s\n",
               strerror(errno));
        free(client);
        pthread_mutexattr_destroy(&attr);
        return VDC_API_ERR_SYSTEM;
    }
    *vdc_client = client;

    return VDC_API_OK;
}

int vdc_api_add_connection(const char *uri)
{
    char *host = NULL;
    char *port = NULL;
    vdc_client_t *vdc_client = NULL;

    if ((!uri) || (strlen(uri) == 0)) {
        vdsm_report(LOG_ERR, "can not add vDC connection: invalid URI\n");
        return VDC_API_ERR_PARAM;
    }

    vdsm_report(LOG_NOTICE, "adding manual vDC connection: %s\n", uri);

    if (!vdsm_util_parse_uri(uri, &host, &port)) {
        vdsm_report(LOG_ERR, "failed to parse uri\n");
        return VDC_API_ERR_PARAM;
    }

    int ret = vdc_api_allocate_client(&vdc_client);
    if (ret != VDC_API_OK) {
        free(host);
        free(port);
        return ret;
    }

    vdc_client->uri = host;
    vdc_client->port = port;

    pthread_rwlock_wrlock(&g_vdc_client_list_mutex);

    LL_PREPEND(g_vdc_client_list, vdc_client);

    pthread_rwlock_unlock(&g_vdc_client_list_mutex);

    return VDC_API_OK;
}

#ifdef HAVE_AVAHI
/* automatically add connection via avahi */
int vdc_api_process_discovered_vdc_host(
    const char *host, const char *port, const dsuid_t *dsuid,
    const char *service_name, bool noauto, bool collection_enabled)
{
    vdc_client_t *vdc_client = NULL;

    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(dsuid, dsuid_str);

    vdsm_report(
        LOG_DEBUG,
        "Considering adding vDC auto-connection to %s:%s / %s / %s\n",
        host, port, dsuid_str, service_name);

    pthread_rwlock_wrlock(&g_vdc_client_list_mutex);

    /* if there is no whitelist, then all devices are allowed, except noauto ones. */
    if (whitelist_exists()) {
        vdc_whitelist_fields_t whitelist_fields;
        whitelist_fields.autoconnect = collection_enabled && !noauto;
        whitelist_fields.noauto = noauto;
        whitelist_fields.name = service_name;
        const bool autoconnect =
            whitelist_insert_and_get_autoconnect(dsuid_str, &whitelist_fields);
        if (!autoconnect) {
            vdsm_report(
                LOG_INFO,
                "Not connecting device %s because it is not whitelisted as autoconnect.\n",
                dsuid_str);
            pthread_rwlock_unlock(&g_vdc_client_list_mutex);
            return VDC_API_OK;
        } else {
            vdsm_report(
                LOG_DEBUG,
                "device %s whitelisted as autoconnect. Connecting...\n",
                dsuid_str);
        }
    } else {
        if (noauto) {
            vdsm_report(
                LOG_INFO,
                "not connecting NOAUTO device %s because there is no whitelist\n",
                dsuid_str);
            pthread_rwlock_unlock(&g_vdc_client_list_mutex);
            return VDC_API_OK;
        }
    }

    /* client may have been already added via a persistent connection, or because
     * the current discovery has been triggered by a service name change, and the
     * the corresponding service removal has not yet arrived.
     */
    LL_FOREACH(g_vdc_client_list, vdc_client) {
        if (strncmp(vdc_client->dsuid, dsuid_str, DSUID_STR_LEN) == 0) {
            if (vdc_client->autodiscovered) {
                if (strcmp(vdc_client->uri, host) != 0 ||
                    strcmp(vdc_client->port, port) != 0) {
                    /* Can this ever happen? Should we rather re-connect here? */
                    vdsm_report(
                        LOG_INFO,
                        "not adding client %s:%s, alreay in the list as %s:%s\n",
                        host, port, vdc_client->uri, vdc_client->port);
                    pthread_rwlock_unlock(&g_vdc_client_list_mutex);
                    return VDC_API_OK;
                }
                if (strcmp(vdc_client->avahi_service_name, service_name) != 0) {
                    /* The service is being renamed */
                    char* old_service_name = vdc_client->avahi_service_name;
                    vdc_client->avahi_service_name = strdup(service_name);
                    if (!vdc_client->avahi_service_name) {
                        vdsm_report(
                            LOG_ERR,
                            "could not allocate memory for new service name\n");
                        vdc_client->avahi_service_name = old_service_name;
                        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
                        return VDC_API_ERR_OUT_OF_MEMORY;
                    }
                    vdsm_report(
                        LOG_DEBUG,
                        "Client %s:%s, alreay in the list. Renamed \"%s\" (was \"%s\")\n",
                        host, port, service_name, old_service_name);
                    free(old_service_name);
                    pthread_rwlock_unlock(&g_vdc_client_list_mutex);
                    return VDC_API_OK;
                }
            }
            vdsm_report(
                LOG_DEBUG,
                "not adding client %s:%s, alreay in the list\n",
                host, port);
            pthread_rwlock_unlock(&g_vdc_client_list_mutex);
            return VDC_API_OK;
        }
    }

    int ret = vdc_api_allocate_client(&vdc_client);
    if (ret != VDC_API_OK) {
        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
        return ret;
    }

    vdc_client->uri = strdup(host);
    if (!vdc_client->uri) {
        vdsm_report(LOG_ERR, "could not allocate memory for service name\n");
        free(vdc_client);
        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
        return VDC_API_ERR_OUT_OF_MEMORY;
    }

    vdc_client->port = strdup(port);
    if (!vdc_client->port) {
        vdsm_report(LOG_ERR, "could not allocate memory for service name\n");
        free(vdc_client->uri);
        free(vdc_client);
        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
        return VDC_API_ERR_OUT_OF_MEMORY;
    }

    vdc_client->autodiscovered = TRUE;
    vdc_client->avahi_service_name = strdup(service_name);
    if (!vdc_client->avahi_service_name) {
        vdsm_report(LOG_ERR, "could not allocate memory for service name\n");
        free(vdc_client->uri);
        free(vdc_client->port);
        free(vdc_client);
        pthread_rwlock_unlock(&g_vdc_client_list_mutex);
        return VDC_API_ERR_OUT_OF_MEMORY;
    }
    strncpy(vdc_client->dsuid, dsuid_str, DSUID_STR_LEN-1);
    vdc_client->dsuid[DSUID_STR_LEN-1] = '\0';

    LL_PREPEND(g_vdc_client_list, vdc_client);

    pthread_rwlock_unlock(&g_vdc_client_list_mutex);

    return VDC_API_OK;
}

void vdc_api_remove_auto_connection(const char *service_name)
{
    vdc_client_t *vdc_client = NULL;
    vdc_client_t *tmp = NULL;

    vdsm_report(LOG_DEBUG, "will try to remove vDC  \"%s\"\n", service_name);

    pthread_rwlock_wrlock(&g_vdc_client_list_mutex);

    LL_FOREACH_SAFE(g_vdc_client_list, vdc_client, tmp) {
        pthread_mutex_lock(&vdc_client->mutex);

        if ((vdc_client->autodiscovered &&
            vdc_client->avahi_service_name &&
            (strcmp(service_name, vdc_client->avahi_service_name) == 0)))
        {
            vdsm_report(LOG_DEBUG, "found vDC %s in the list, removing it\n",
                   service_name);

            // remove client from devices
            vdc_api_remove_devices_from_vdc_lookup_tables(vdc_client);

            LL_DELETE(g_vdc_client_list, vdc_client);
            if (vdc_client->fd != -1) {
                close(vdc_client->fd);
            }

            if (vdc_client->uri) {
                free(vdc_client->uri);
            }

            if (vdc_client->port) {
                free(vdc_client->port);
            }

            if (vdc_client->avahi_service_name) {
                free(vdc_client->avahi_service_name);
            }

            pthread_mutex_unlock(&vdc_client->mutex);
            pthread_mutex_destroy(&vdc_client->mutex);
            free(vdc_client);
            break;
        }

        pthread_mutex_unlock(&vdc_client->mutex);
    }

    pthread_rwlock_unlock(&g_vdc_client_list_mutex);
}

bool vdc_api_in_client_list(const dsuid_t* dsuid)
{
    vdc_client_t *vdc_client = NULL;
    bool found = false;
    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(dsuid, dsuid_str);

    pthread_rwlock_rdlock(&g_vdc_client_list_mutex);

    LL_FOREACH(g_vdc_client_list, vdc_client) {
        if (strncmp(vdc_client->dsuid, dsuid_str, DSUID_STR_LEN) == 0) {
            found = true;
            break;
        }
    }

    pthread_rwlock_unlock(&g_vdc_client_list_mutex);

    return found;
}

void vdc_api_whitelist_changed()
{
    g_whitelist_changed = true;
}

#endif//HAVE_AVAHI

void vdc_api_shutdown()
{
    ++g_shutdown_flag;
}

bool vdc_is_down()
{
    return g_vdc_down;
}

void vdc_api_user_protobuf_message_request(uint16_t size, const uint8_t *request, const dsuid_t *dsuid)
{
    vdsm_report(LOG_DEBUG, "vdc api user protobuf message size: %d\n", size);

    char vdc_str[DSUID_STR_LEN];
    dsuid_to_string_upper(dsuid, vdc_str);

    Vdcapi__Message *msg = vdcapi__message__unpack(NULL, size, request);
    if (!msg) {
        vdsm_report(LOG_ERR, "vdc api user protobuf message: failed to unpack incoming message\n");
        return;
    }

    request_cache_t *cache;
    cache = (request_cache_t*) malloc(sizeof(request_cache_t));
    if (!cache) {
        vdc_api_free_message(msg);
        vdsm_report(LOG_ERR, "could not allocate memory for protobuf message request cache\n");
        return;
    }
    cache->context = REQUEST_CONTEXT_GET_PROTOBUF_DATA;

    if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_vdc_dsuid(vdc_str, msg, cache)) {
        // free message only, if it was not injected in vdc_api_send_message
        // otherwise message will be auto removed after timeout.
        vdsm_report(LOG_ERR,
            "FAILED: protobuf message request: for vdc  %s\n", vdc_str);
        vdc_api_free_message(msg);
        free(cache);
    }
}

int vdc_api_init(const dsuid_t *dsmid, VDSMRole role, uint16_t listen_port)
{
    g_vdsm_listen_port = listen_port;
    g_vdsm_role = role;
    g_dsm_dsuid = *dsmid;

    /* init mutexes */

    pthread_mutexattr_t attr;
    if (pthread_mutexattr_init(&attr) != 0) {
        vdsm_report(LOG_ERR, "failed to init mutex attribute: %s\n",
               strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0) {
        pthread_mutexattr_destroy(&attr);
        vdsm_report(LOG_ERR, "failed to set mutex type: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_rwlock_init(&g_vdc_client_list_mutex, NULL) != 0) {
        pthread_mutexattr_destroy(&attr);
        vdsm_report(LOG_ERR, "failed init mutex: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_mutex_init(&g_vdc_device_lookup_hash_mutex, &attr) != 0) {
        pthread_mutexattr_destroy(&attr);
        pthread_rwlock_destroy(&g_vdc_client_list_mutex);
        vdsm_report(LOG_ERR, "failed init mutex: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_mutex_init(&g_cached_request_list_mutex, NULL) != 0) {
        pthread_mutexattr_destroy(&attr);
        pthread_rwlock_destroy(&g_vdc_client_list_mutex);
        pthread_mutex_destroy(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "failed init mutex: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    if (pthread_mutex_init(&g_vdc_lookup_hash_mutex, &attr) != 0) {
        pthread_mutexattr_destroy(&attr);
        pthread_rwlock_destroy(&g_vdc_client_list_mutex);
        pthread_mutex_destroy(&g_vdc_device_lookup_hash_mutex);
        pthread_mutex_destroy(&g_cached_request_list_mutex);
        vdsm_report(LOG_ERR, "failed init mutex: %s\n", strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    pthread_mutexattr_destroy(&attr);

    if (pthread_create(&g_vdc_api_thread, NULL, vdc_api_service, NULL) != 0) {
        pthread_mutexattr_destroy(&attr);
        pthread_rwlock_destroy(&g_vdc_client_list_mutex);
        pthread_mutex_destroy(&g_cached_request_list_mutex);
        pthread_mutex_destroy(&g_vdc_device_lookup_hash_mutex);
        pthread_mutex_destroy(&g_vdc_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "failed to launch vDC service thread: %s\n",
               strerror(errno));
        return VDC_API_ERR_SYSTEM;
    }

    /* TODO: subscribe to int/term signals and do not detach the thread,
     * implement a clean shutodown and join it */
    if (pthread_detach(g_vdc_api_thread) != 0) {
        g_shutdown_flag++;
        pthread_mutexattr_destroy(&attr);
        pthread_rwlock_destroy(&g_vdc_client_list_mutex);
        pthread_mutex_destroy(&g_cached_request_list_mutex);
        pthread_mutex_destroy(&g_vdc_device_lookup_hash_mutex);
        pthread_mutex_destroy(&g_vdc_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "failed to detach vDC service thread: %s\n",
               strerror(errno));
    }

    return VDC_API_OK;
}

int vdc_api_send_message_by_device_dsuid(const char *dev_dsuid, Vdcapi__Message *msg,
                          request_cache_t *rc)
{
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (vdc_api_get_vdc_from_lookup_table_private(dev_dsuid, &vdc_client)) {
        int retVal = vdc_api_send_message(vdc_client, msg, rc);
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        return retVal;
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "vdc api send message: client %s not found.\n", dev_dsuid);
    }
    return VDC_API_ERR_CLIENT_NOT_FOUND;
}

int vdc_api_send_message_by_vdc_dsuid(const char *vdc_dsuid,
                                      Vdcapi__Message *msg,
                                      request_cache_t *rc)
{
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);
    if (vdc_api_get_vdc_host_for_vdc_private(vdc_dsuid, &vdc_client)) {
        int retVal = vdc_api_send_message(vdc_client, msg, rc);
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
        return retVal;
    } else {
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "vdc api send message: client %s not found.\n", vdc_dsuid);
    }
    return VDC_API_ERR_CLIENT_NOT_FOUND;
}

int vdc_api_send_message(vdc_client_t *vdc_client, Vdcapi__Message *msg,
                         request_cache_t *rc)
{
    vdsm_report(LOG_DEBUG, "Send message \n");

    pthread_mutex_lock(&vdc_client->mutex);
    if (vdc_client->fd == -1)
    {
        pthread_mutex_unlock(&vdc_client->mutex);
        return VDC_API_ERR_NOT_CONNECTED;
    }

    if (vdc_api_msg_is_request(msg->type)) {
        g_message_id++;
        if (g_message_id == PROTO_MESSAGE_ID_INIT) {
            g_message_id++;
        }
        msg->message_id = g_message_id;
        msg->has_message_id = 1;
        int ret = vdc_api_cache_request(msg, rc);
        if (ret != VDC_API_OK) {
            pthread_mutex_unlock(&vdc_client->mutex);
            return ret;
        }
    } else {
        /* request context makes no sense with messages that are not cached */
        if (rc) {
            vdsm_report(LOG_DEBUG, "context cache given for a non cached message\n");
            pthread_mutex_unlock(&vdc_client->mutex);
            return VDC_API_ERR_PARAM;
        }
    }

    vdsm_report(LOG_DEBUG, "sending message to vDC (msg type): %d\n", msg->type);

    size_t msg_len = vdcapi__message__get_packed_size(msg);
    uint8_t *msg_buf = malloc(msg_len * sizeof(uint8_t));
    if (!msg_buf)
    {
        vdsm_report(LOG_ERR, "could not allocate %zu bytes for protobuf message "
                        "serialization.\n", msg_len);
        pthread_mutex_unlock(&vdc_client->mutex);
        return VDC_API_ERR_OUT_OF_MEMORY;
    }

    vdcapi__message__pack(msg, msg_buf);

/* TODO temp to get message
    size_t j = 0;
    fprintf( stderr, "cmd = [");
    for (j = 0; j < msg_len-1; ++j)
    {
        fprintf( stderr, "0x%02x, ", msg_buf[j]);
    }
    fprintf( stderr, "0x%02x ]\n", msg_buf[msg_len-1]);
 TODO END */

    /* log protobuf message */
    protobufMessagePrint(&msg->base,0);

    uint16_t netlen = htons(msg_len);
    if (sockwrite(vdc_client->fd, (unsigned char *)&netlen,
                sizeof(uint16_t), VDC_SOCKET_WRITE_TIMEOUT) != sizeof(uint16_t))
    {
        vdsm_report(LOG_ERR, "could not send message length to vdC host %s\n",
               vdc_client->dsuid);
        free(msg_buf);
        pthread_mutex_unlock(&vdc_client->mutex);
        return VDC_API_ERR_SOCKET;
    }

    int written = sockwrite(vdc_client->fd, (unsigned char *)msg_buf, msg_len,
                            VDC_SOCKET_WRITE_TIMEOUT);
    if ((written < 0) || ((size_t)written != msg_len))
    {
        vdsm_report(LOG_ERR, "could not send message to vdC host %s\n",
               vdc_client->dsuid);
        free(msg_buf);
        pthread_mutex_unlock(&vdc_client->mutex);
        return VDC_API_ERR_SOCKET;
    }

    free(msg_buf);
    pthread_mutex_unlock(&vdc_client->mutex);
    return VDC_API_OK;
}

void vdc_api_send_get_property_by_device_dsuid(const char *dsuid,
                                        const char *name,
                                        request_cache_t *rc)
{
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client)) {
        vdc_api_send_get_property(vdc_client, dsuid, name, rc);
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        free (rc);
        vdsm_report(LOG_ERR, "vdc api send get property by dsuid :"
                    " client %s not found.\n", dsuid);
    }
}

void vdc_api_send_get_property_by_vdc_dsuid(const char *vdc_dsuid,
                                            const char *name,
                                            request_cache_t *rc)
{
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);
    if (vdc_api_get_vdc_host_for_vdc_private(vdc_dsuid, &vdc_client)) {
        vdc_api_send_get_property(vdc_client, vdc_dsuid, name, rc);
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
        free (rc);
        vdsm_report(LOG_ERR, "vdc api send get property by vdc dsuid %s\n", vdc_dsuid);
    }
}

void vdc_api_send_get_property(vdc_client_t *vdc_client, const char *dsuid,
                               const char *name,
                               request_cache_t *rc)
{
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestGetProperty *getprop_submsg = NULL;
    Vdcapi__PropertyElement* elemPropName = NULL;
    Vdcapi__PropertyElement* elemPropIndex = NULL;

    if (!dsuid) {
        vdsm_report(LOG_ERR, "dsuid does not exist\n");
        goto cleanup_get_property;
    }

    msg = malloc(sizeof(Vdcapi__Message));
    if (!msg) {
        vdsm_report(LOG_ERR, "failed to allocate memory for message VdsmRequestGetProperty.\n");
        goto cleanup_get_property;
    }
    vdcapi__message__init(msg);


    if (proputil_get_property_new(dsuid, &getprop_submsg) != 0) {
        goto cleanup_get_property;
    }


    if (proputil_create_property_empty(&elemPropName, name) != 0) {
        goto cleanup_get_property;
    }
    if (proputil_get_property_add(getprop_submsg, elemPropName) !=0) {
        proputil_property_recursive_free(elemPropName);
        goto cleanup_get_property;
    }

    if (proputil_create_property_empty(&elemPropIndex, "") != 0) {
        goto cleanup_get_property;
    }
    if (proputil_add_property_element(elemPropName, elemPropIndex) != 0) {
        proputil_property_recursive_free(elemPropIndex);
        goto cleanup_get_property;
    }

    msg->vdsm_request_get_property = getprop_submsg;
    msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;

    // message must be allocated because this request is stored
    // it will be freed once it is no longer needed
    vdc_api_send_message(vdc_client, msg, rc);
    return;

cleanup_get_property:
    vdsm_report(LOG_ERR, "vdc api send get property: failed to allocate memory for elements.\n");
    free(rc);
    free(msg);
    proputil_get_property_free(getprop_submsg);
}

void vdc_api_send_get_properties_indexed_by_device_dsuid(const char *dsuid,
                                                  const char *name,
                                                  uint32_t* array,
                                                  uint32_t array_len,
                                                  request_cache_t *rc)
{
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client)) {
        vdc_api_send_get_properties_indexed(vdc_client, dsuid, name, array, array_len, rc);
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        free (rc);
        vdsm_report(LOG_ERR, "vdc api send get property indexed by dsuid :"
                    "client %s not found.\n", dsuid);
    }
}

void vdc_api_send_get_properties_indexed(vdc_client_t *vdc_client,
                                         const char *dsuid,
                                         const char *name,
                                         uint32_t* array,
                                         uint32_t array_len,
                                         request_cache_t *rc)
{
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestGetProperty *getprop_submsg = NULL;
    Vdcapi__PropertyElement* elemPropName = NULL;

    if (!dsuid) {
        vdsm_report(LOG_ERR, "dsuid does not exist\n");
        return;
    }

    msg = malloc(sizeof(Vdcapi__Message));
    if (!msg) {
        vdsm_report(LOG_ERR, "failed to allocate memory for message VdsmRequestGetProperty.\n");
        return;
    }
    vdcapi__message__init(msg);

    if (proputil_get_property_new(dsuid, &getprop_submsg) != 0) {
        goto cleanup_get_roperties_indexed;
    }

    if (proputil_create_property_empty(&elemPropName, name) != 0) {
        goto cleanup_get_roperties_indexed;
    }

    if (proputil_get_property_add(getprop_submsg, elemPropName) !=0) {
        proputil_property_recursive_free(elemPropName);
        goto cleanup_get_roperties_indexed;
    }

    size_t i;
    for (i = 0; i < array_len; ++i) {
        Vdcapi__PropertyElement* elemPropIndex = NULL;
        if (proputil_create_property_empty_index(&elemPropIndex, array[i]) != 0) {
            goto cleanup_get_roperties_indexed;
        }
        if (proputil_add_property_element(elemPropName, elemPropIndex) != 0) {
            proputil_property_recursive_free(elemPropIndex);
            goto cleanup_get_roperties_indexed;
        }
    }

    msg->vdsm_request_get_property = getprop_submsg;
    msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;

    // message must be allocated because this request is stored
    // it will be freed once it is no longer needed
    vdc_api_send_message(vdc_client, msg, rc);
    return;

cleanup_get_roperties_indexed:
    vdsm_report(LOG_ERR, "vdc api send get properties indexed: failed to allocate memory for elements.\n");
    free(rc);
    free(msg);
    proputil_get_property_free(getprop_submsg);
}

void vdc_api_device_get_name(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device get name). device: %d\n", device_id);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        request_cache_t *cache;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL)
        {
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                            "cache\n");
            return;
        }
        cache->context = REQUEST_CONTEXT_GET_PROPERTY_NAME;
        vdc_api_send_get_property_by_device_dsuid(dsuid, "name", cache);
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device get name). device: %d not found\n",
               device_id);
    }
}

void vdc_api_device_set_name(uint16_t device_id, const char *name)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device set name). device: %d, name: %s\n",
           device_id, name);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        Vdcapi__Message *msg;
        Vdcapi__VdsmRequestSetProperty *submsg;

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            vdsm_report(LOG_ERR, "vdc api request set property: failed to allocate "
                            "memory\n");
            return;
        }
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
             free(msg);
             return;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        Vdcapi__PropertyElement* propElem = NULL;
        if (proputil_set_property_add_string(submsg, "name", name, &propElem) != 0) {
            free(msg);
            proputil_set_property_free(submsg);
            return;
        }

        // message will be freed when response is received
        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, NULL)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device set name). device: %d not found\n", device_id);
    }
}

void vdc_api_device_register_sensors(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device register sensors). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid)) {
        request_cache_t *cache;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL)
        {
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                            "cache\n");
            return;
        }
        cache->context = REQUEST_CONTEXT_GET_PROPERTY_SENSORS;
        vdc_api_send_get_property_by_device_dsuid(dsuid, "sensorDescriptions", cache);
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device registers sensors). device: %d "
                            "not found\n", device_id);
    }
}

void vdc_api_device_register_opc_channels(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device get opc channels). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        request_cache_t *cache = NULL;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (!cache) {
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                            "cache\n");
            return;
        }
        cache->context = REQUEST_CONTEXT_GET_PROPERTY_OPC_VALUES;
        cache->count  = 0;

        vdc_api_send_get_property_by_device_dsuid(dsuid, "channelDescriptions", cache);
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device get opc values). device: %d "
                            "not found\n", device_id);
    }
}

void vdc_api_device_register_binary_inputs(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device get binary inputs). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        request_cache_t *cache;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL)
        {
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                    "cache\n");
            return;
        }

        cache->context = REQUEST_CONTEXT_GET_PROPERTY_BINARY_INPUTS;
        cache->count = 0;
        vdc_api_send_get_property_by_device_dsuid(dsuid, "binaryInputDescriptions", cache);
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device get binary inputs). device: %d "
                            "not found\n", device_id);
    }
}

void vdc_api_device_get_group_membership(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];

    vdsm_report(LOG_DEBUG, "vdc api (device get groupMembership). device: %d\n",
           device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        request_cache_t *cache;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL)
        {
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                            "cache\n");
            return;
        }

        cache->context = REQUEST_CONTEXT_GET_PROPERTY_GROUP_MEMBERSHIP;
        cache->result = device_id;

        vdc_api_send_get_property_by_device_dsuid(dsuid, "outputSettings", cache);
    } else {
         vdsm_report(LOG_WARNING, "vdc api (device get groupMembership). device: %d "
                             "not found\n", device_id);
    }
}

void vdc_api_device_ping(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device ping). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        Vdcapi__Message msg = VDCAPI__MESSAGE__INIT;
        Vdcapi__VdsmSendPing submsg =
            VDCAPI__VDSM__SEND_PING__INIT;

        submsg.dsuid = dsuid;

        msg.type = VDCAPI__TYPE__VDSM_SEND_PING;
        msg.vdsm_send_ping = &submsg;

        if (VDC_API_OK != vdc_api_send_message_by_device_dsuid(dsuid, &msg, NULL)) {
            vdsm_report(LOG_WARNING, "vdc api (device ping). device: %d send message error\n", device_id);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device ping). device: %d not found\n", device_id);
    }
}

void vdc_api_device_remove(uint16_t device_id)
{
    char dsuid[DSUID_STR_LEN];
    vdsm_report(LOG_DEBUG, "vdc api (device remove). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        request_cache_t *cache;
        Vdcapi__Message *msg;
        Vdcapi__VdsmSendRemove *remove_submsg;

        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL) {
            vdsm_report(LOG_ERR, "could not allocate memory for remove request "
                                 "cache\n");
            return;
        }
        msg = malloc(sizeof(Vdcapi__Message));
        if (msg == NULL) {
            vdsm_report(LOG_ERR, "vdc api send remove: failed to allocate memory.\n");
            free(cache);
            return;
        }
        remove_submsg = malloc(sizeof(Vdcapi__VdsmSendRemove));
        if (remove_submsg == NULL) {
            vdsm_report(LOG_ERR, "vdc api send remove: failed to allocate memory.\n");
            free(msg);
            free(cache);
            return;
        }
        vdcapi__message__init(msg);
        vdcapi__vdsm__send_remove__init(remove_submsg);

        remove_submsg->dsuid = dsuid;

        msg->type = VDCAPI__TYPE__VDSM_SEND_REMOVE;
        msg->vdsm_send_remove = remove_submsg;

        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, cache)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
            free(cache);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device remove). device: %d not found\n", device_id);
    }
}

//******************************************************************************
void vdc_api_device_set_opc_output_state(uint16_t device_id, uint8_t channel,
                                         uint16_t value, bool apply)
{
    char dsuid[DSUID_STR_LEN];

    vdsm_report(LOG_DEBUG, "vdc api (device set opc parameter). device: %d\n", device_id);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        double output = 0;
        if (!convert_dsmval_to_vdcapival(channel, value, &output)) {
            vdsm_report(LOG_ERR, "vdc api request set value: failed get conversion parameters\n");
            return;
        }

        Vdcapi__Message msg = VDCAPI__MESSAGE__INIT;
        Vdcapi__VdsmNotificationSetOutputChannelValue submsg =
                VDCAPI__VDSM__NOTIFICATION_SET_OUTPUT_CHANNEL_VALUE__INIT;

        char *dsuids[1];
        dsuids[0] = (char *) dsuid;
        submsg.dsuid = dsuids;
        submsg.n_dsuid = 1;

        msg.vdsm_send_output_channel_value = &submsg;
        msg.type = VDCAPI__TYPE__VDSM_NOTIFICATION_SET_OUTPUT_CHANNEL_VALUE;

        submsg.apply_now = apply;
        submsg.has_apply_now = TRUE;

        submsg.channel = channel;
        submsg.has_channel = TRUE;

        submsg.value = output;
        submsg.has_value = TRUE;

        // static message, not stored for response tracking
        if (VDC_API_OK != vdc_api_send_message_by_device_dsuid(dsuid, &msg,  NULL)) {
            vdsm_report(LOG_WARNING, "vdc api (set opc output state. device: %d "
                          "message not sent\n", device_id);
        }
    } else {
      vdsm_report(LOG_WARNING, "vdc api (set opc output state. device: %d "
                          "not found\n", device_id);
    }
 }

//******************************************************************************
void vdc_api_device_get_opc_output_state(uint16_t device_id, uint8_t channel)
{
    char dsuid[DSUID_STR_LEN];

    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestGetProperty *submsg = NULL;
    request_cache_t *cache = NULL;
    Vdcapi__PropertyElement* elemChannelStates = NULL;
    Vdcapi__PropertyElement* elemChannelId = NULL;
    Vdcapi__PropertyElement* elemValue = NULL;

    vdsm_report(LOG_DEBUG, "vdc api (device get output state). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            goto cleanup_get_opc_output_state;
        }
        vdcapi__message__init(msg);

        if (proputil_get_property_new(dsuid, &submsg) != 0) {
            goto cleanup_get_opc_output_state;
        }

        msg->vdsm_request_get_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;

        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (!cache) goto cleanup_get_opc_output_state;
        cache->context = REQUEST_CONTEXT_GET_PROPERTY_OUTPUT_VALUE;
        cache->channel = channel;

        // channelStates
        if (proputil_create_property_empty(&elemChannelStates,"channelStates") != 0) {
            goto cleanup_get_opc_output_state;
        }
        if (proputil_get_property_add(submsg,elemChannelStates) != 0) {
            proputil_property_recursive_free(elemChannelStates);
            goto cleanup_get_opc_output_state;
        }

        //ChannelId
        if (proputil_create_property_empty_index(&elemChannelId, channel) != 0) {
            goto cleanup_get_opc_output_state;
        }
        if (proputil_add_property_element(elemChannelStates, elemChannelId) != 0) {
            proputil_property_recursive_free(elemChannelId);
            goto cleanup_get_opc_output_state;
        }

        // Value
        if (proputil_create_property_empty(&elemValue, "value") != 0) {
            goto cleanup_get_opc_output_state;
        }
        if (proputil_add_property_element(elemChannelId, elemValue) != 0) {
            proputil_property_recursive_free(elemValue);
            goto cleanup_get_opc_output_state;
        }

        // message will be freed when response is received
        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, cache)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
            free(cache);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (get opc parameter). device: %d "
                         "not found\n", device_id);
    }
    return;

cleanup_get_opc_output_state:
    vdsm_report(LOG_ERR, "vdc api send get property: failed to allocate memory.\n");
    free(msg);
    proputil_get_property_free(submsg);
    free(cache);
}

//******************************************************************************
void vdc_api_device_set_opc_scene(uint16_t device_id, uint8_t OPCscene,
                                  uint8_t channel, uint16_t value)
{
    char dsuid[DSUID_STR_LEN];
    Vdcapi__Message *msg;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;
    Vdcapi__PropertyElement* elemScene = NULL;
    Vdcapi__PropertyElement* elemSceneId = NULL;
    Vdcapi__PropertyElement* elemChannel = NULL;
    Vdcapi__PropertyElement* elemChannelId = NULL;
    Vdcapi__PropertyElement* elemValue = NULL;

    vdsm_report(LOG_DEBUG, "vdc api (device set opc scene). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        double outputValue = 0;
        if (!convert_dsmval_to_vdcapival(channel, value, &outputValue)) {
            vdsm_report(LOG_ERR, "vdc api set opc scene value: failed get conversion parameters\n");
            return;
        }

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) goto cleanup_set_opc_scene;
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
            goto cleanup_set_opc_scene;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        // Scene
        if (proputil_create_property_empty(&elemScene, "scenes") != 0) {
            goto cleanup_set_opc_scene;
        }
        if (proputil_set_property_add(submsg, elemScene) != 0) {
            proputil_property_recursive_free(elemScene);
            goto cleanup_set_opc_scene;
        }

        // SceneId
        if (proputil_create_property_empty_index(&elemSceneId, OPCscene) != 0) {
            goto cleanup_set_opc_scene;
        }
        if (proputil_add_property_element(elemScene, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_set_opc_scene;
        }

        // Channels
        if (proputil_create_property_empty(&elemChannel, "channels") != 0) {
            goto cleanup_set_opc_scene;
        }
        if (proputil_add_property_element(elemSceneId, elemChannel) != 0) {
            proputil_property_recursive_free(elemChannel);
            goto cleanup_set_opc_scene;
        }

        // ChannelId
        if (proputil_create_property_empty_index(&elemChannelId, channel) != 0) {
            goto cleanup_set_opc_scene;
        }
        if (proputil_add_property_element(elemChannel, elemChannelId) != 0) {
            proputil_property_recursive_free(elemChannelId);
            goto cleanup_set_opc_scene;
        }

        // Value
        if (proputil_create_property_double(&elemValue, "value", outputValue) != 0) {
            goto cleanup_set_opc_scene;
        }
        if (proputil_add_property_element(elemChannelId, elemValue) != 0) {
            proputil_property_recursive_free(elemValue);
            goto cleanup_set_opc_scene;
        }

        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, NULL)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (set opc scene). device: %d "
                         "not found\n", device_id);
    }
    return;

cleanup_set_opc_scene:
    vdsm_report(LOG_ERR, "vdc api send set scene property: failed to allocate memory.\n");
    free(msg);
    proputil_set_property_free(submsg);
}

/*
 * Bank 103
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
bool vdc_api_device_set_opc_scene_config(uint16_t device_id, uint8_t OPCscene,
                                         uint16_t value)
{
    char dsuid[DSUID_STR_LEN];
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;

    Vdcapi__PropertyElement *elemScene = NULL;
    Vdcapi__PropertyElement *elemSceneId = NULL;
    Vdcapi__PropertyElement *elemIgnoreLocalPriority = NULL;
    Vdcapi__PropertyElement *elemDontCare = NULL;
    Vdcapi__PropertyElement *elemEffect = NULL;
    bool success = FALSE;

    vdsm_report(LOG_DEBUG, "vdc api (device set opc scene config). device: %d\n", device_id);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        // Evaluate config parameters from bank 103
        bool dontCare = (bool)(value&0x01);
        bool ignoreLocalPriority = (bool)(value&0x02);
        uint64_t dsseffect = (uint64_t)((value>>8)&0x0f);
        // at this time, dSS only knows effect==0 (no effect) and effect==1 (blink)
        // This must be mapped to vDC API effects 1 (standard transition) and 4 (blink), resp.
        uint64_t vdceffect = dsseffect==1 ? 4 : 1;

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) goto cleanup_set_opc_scene_config;
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
            goto cleanup_set_opc_scene_config;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        // Scenes
        if (proputil_create_property_empty(&elemScene, "scenes") != 0) {
            goto cleanup_set_opc_scene_config;
        }
        if (proputil_set_property_add(submsg, elemScene) != 0) {
            proputil_property_recursive_free(elemScene);
            goto cleanup_set_opc_scene_config;
        }

        // SceneId
        if (proputil_create_property_empty_index(&elemSceneId, OPCscene) != 0) {
            goto cleanup_set_opc_scene_config;
        }
        if (proputil_add_property_element(elemScene, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_set_opc_scene_config;
        }

        // dontCare
        if (proputil_create_property_bool(&elemDontCare, "dontCare", dontCare) != 0) {
            goto cleanup_set_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemDontCare) != 0) {
            proputil_property_recursive_free(elemDontCare);
            goto cleanup_set_opc_scene_config;
        }

        // ignoreLocalPriority
        if (proputil_create_property_bool(&elemIgnoreLocalPriority, "ignoreLocalPriority", ignoreLocalPriority) != 0) {
            goto cleanup_set_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemIgnoreLocalPriority) != 0) {
            proputil_property_recursive_free(elemIgnoreLocalPriority);
            goto cleanup_set_opc_scene_config;
        }

        if (proputil_create_property_uint(&elemEffect, "effect", vdceffect) != 0) {
            goto cleanup_set_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemEffect) != 0) {
            proputil_property_recursive_free(elemEffect);
            goto cleanup_set_opc_scene_config;
        }

        // message will be freed when response is received
        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, NULL)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
        } else {
            success = TRUE;
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (set opc scene). device: %d not found\n", device_id);
    }
    return success;

cleanup_set_opc_scene_config:
    vdsm_report(LOG_ERR, "vdc api send set opc scene config: failed to allocate memory.\n");
    vdc_api_free_message(msg);
    return FALSE;
}

/*
 * Bank 104
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
bool vdc_api_device_set_opc_scene_dc_flags(uint16_t device_id, uint8_t OPCscene,
                                           uint16_t value)
{
    char dsuid[DSUID_STR_LEN];
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;
    Vdcapi__PropertyElement* elemScenes = NULL;
    Vdcapi__PropertyElement* elemSceneId = NULL;
    Vdcapi__PropertyElement* elemChannels = NULL;
    bool success = FALSE;

    vdsm_report(LOG_DEBUG, "vdc api (device set opc scene dc flags). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            goto cleanup_set_opc_scene_dc_flags;
        }
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
            goto cleanup_set_opc_scene_dc_flags;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        // Scene
        if (proputil_create_property_empty(&elemScenes, "scenes") != 0) {
            goto cleanup_set_opc_scene_dc_flags;
        }
        if (proputil_set_property_add(submsg, elemScenes) != 0) {
            proputil_property_recursive_free(elemScenes);
            goto cleanup_set_opc_scene_dc_flags;
        }

        // Scene Id
        if (proputil_create_property_empty_index(&elemSceneId, OPCscene) != 0) {
            goto cleanup_set_opc_scene_dc_flags;
        }
        if (proputil_add_property_element(elemScenes, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_set_opc_scene_dc_flags;
        }

        // TODO Channels
        if (proputil_create_property_empty(&elemChannels, "channels") != 0) {
            goto cleanup_set_opc_scene_dc_flags;
        }
        if (proputil_add_property_element(elemSceneId, elemChannels) != 0) {
            proputil_property_recursive_free(elemChannels);
            goto cleanup_set_opc_scene_dc_flags;
        }

        int i;
        for (i = 0; i < 16; ++i) {
            Vdcapi__PropertyElement* elemChannelId = NULL;
            Vdcapi__PropertyElement* elemDontCare = NULL;

            bool dontCare = (bool)((value >> i)&0x01);

            uint8_t channelId;
            if (!Datamodel_GetOpcChannelFromIndex(device_id,i,&channelId)) continue;

            if (proputil_create_property_empty_index(&elemChannelId, channelId) != 0) {
                goto cleanup_set_opc_scene_dc_flags;
            }
            if (proputil_add_property_element(elemChannels, elemChannelId) != 0) {
                proputil_property_recursive_free(elemChannelId);
                goto cleanup_set_opc_scene_dc_flags;
            }

            if (proputil_create_property_bool(&elemDontCare, "dontCare", dontCare) != 0) {
                goto cleanup_set_opc_scene_dc_flags;
            }
            if (proputil_add_property_element(elemChannelId, elemDontCare) != 0) {
                proputil_property_recursive_free(elemDontCare);
                goto cleanup_set_opc_scene_dc_flags;
            }
        }

        // message will be freed when response is received
        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, NULL)) {
            // free message only, if it was not injected in vdc_api_send_message
            // otherwise message will be auto removed after timeout.
            vdc_api_free_message(msg);
        } else {
            success = TRUE;
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (set opc scene). device: %d not found\n", device_id);
    }
    return success;

cleanup_set_opc_scene_dc_flags:
    vdsm_report(LOG_ERR, "vdc api request get property: failed to allocate memory\n");
    free(msg);
    proputil_set_property_free(submsg);
    return FALSE;
}

void vdc_api_device_get_opc_scene_value(uint16_t device_id, uint8_t OPCscene, uint8_t channel)
{
    char dsuid[DSUID_STR_LEN];
    request_cache_t *cache = NULL;
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestGetProperty *getprop_submsg = NULL;
    Vdcapi__PropertyElement* elemScene = NULL;
    Vdcapi__PropertyElement* elemSceneId = NULL;
    Vdcapi__PropertyElement* elemChannel = NULL;
    Vdcapi__PropertyElement* elemChannelId = NULL;
    Vdcapi__PropertyElement* elemValue = NULL;

    vdsm_report(LOG_DEBUG, "vdc api (device get opc scene value). device: %d\n", device_id);

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (!cache) {
            vdsm_report(LOG_ERR, "could not allocate memory for property request cache\n");
            return;
        }

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            goto cleanup_get_opc_scene_value;
        }
        vdcapi__message__init(msg);

        if (proputil_get_property_new(dsuid, &getprop_submsg) != 0) {
            goto cleanup_get_opc_scene_value;
        }
        msg->vdsm_request_get_property = getprop_submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;

        // Scenes
        if (proputil_create_property_empty(&elemScene, "scenes") != 0) {
            goto cleanup_get_opc_scene_value;
        }
        if (proputil_get_property_add(getprop_submsg, elemScene) != 0) {
            proputil_property_recursive_free(elemScene);
            goto cleanup_get_opc_scene_value;
        }

        // Scene id
        if (proputil_create_property_empty_index(&elemSceneId, OPCscene) != 0) {
            goto cleanup_get_opc_scene_value;
        }
        if (proputil_add_property_element(elemScene, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_get_opc_scene_value;
        }

        // Channels
        if (proputil_create_property_empty(&elemChannel, "channels") != 0) {
            goto cleanup_get_opc_scene_value;
        }
        if (proputil_add_property_element(elemSceneId, elemChannel) != 0) {
            proputil_property_recursive_free(elemChannel);
            goto cleanup_get_opc_scene_value;
        }

        // Channel Id
        if (proputil_create_property_empty_index(&elemChannelId, channel) != 0) {
            goto cleanup_get_opc_scene_value;
        }
        if (proputil_add_property_element(elemChannel, elemChannelId) != 0) {
            proputil_property_recursive_free(elemChannelId);
            goto cleanup_get_opc_scene_value;
        }

        // Value
        if (proputil_create_property_empty(&elemValue, "value") != 0) {
            goto cleanup_get_opc_scene_value;
        }
        if (proputil_add_property_element(elemChannelId, elemValue) != 0) {
            proputil_property_recursive_free(elemValue);
            goto cleanup_get_opc_scene_value;
        }

        cache->context = REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENES_VALUE;
        cache->channel = channel;
        cache->scene  = OPCscene;

        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, cache)) {
          // free message only, if it was not injected in vdc_api_send_message
          // otherwise message will be auto removed after timeout.
          vdc_api_free_message(msg);
          free(cache);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (set opc scene). device: %d "
                            "not found\n", device_id);
    }
    return;

cleanup_get_opc_scene_value:
    vdsm_report(LOG_ERR, "vdc api vdc api (device get opc scene value): failed to allocate memory for elements.\n");
    free(msg);
    free(cache);
    proputil_get_property_free(getprop_submsg);
}

void vdc_api_device_get_opc_scene_config(uint16_t device_id, uint8_t OPCscene, uint8_t OPCoperation)
{
    char dsuid[DSUID_STR_LEN];
    request_cache_t *cache;
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestGetProperty *submsg = NULL;

    Vdcapi__PropertyElement* elemScene = NULL;
    Vdcapi__PropertyElement* elemSceneId = NULL;
    Vdcapi__PropertyElement* elemOutput = NULL;
    Vdcapi__PropertyElement* elemOutputId = NULL;

    Vdcapi__PropertyElement* elemChannelDontcare = NULL;
    Vdcapi__PropertyElement* elemIgnoreLocalPriority = NULL;
    Vdcapi__PropertyElement* elemTransitionEffect = NULL;
    Vdcapi__PropertyElement* elemSceneDontCare = NULL;

    vdsm_report(LOG_DEBUG, "vdc api (device get opc scene). device: %d\n", device_id);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (!cache) {
            vdsm_report(LOG_ERR, "could not allocate memory for property request cache\n");
            return;
        }
        cache->operation = OPCoperation;
        cache->context = REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENE_CONFIG;

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            goto cleanup_get_opc_scene_config;
        }
        vdcapi__message__init(msg);

        if (proputil_get_property_new(dsuid, &submsg) != 0) {
            goto cleanup_get_opc_scene_config;
        }
        msg->vdsm_request_get_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;

        // Scenes
        if (proputil_create_property_empty(&elemScene, "scenes") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_get_property_add(submsg, elemScene) != 0) {
            proputil_property_recursive_free(elemScene);
            goto cleanup_get_opc_scene_config;
        }

        // SceneId
        if (proputil_create_property_empty_index(&elemSceneId, OPCscene) != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemScene, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_get_opc_scene_config;
        }

        // Channel
        if (proputil_create_property_empty(&elemOutput, "channels") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemOutput) != 0) {
            proputil_property_recursive_free(elemOutput);
            goto cleanup_get_opc_scene_config;
        }

        // OutputId -> all
        if (proputil_create_property_empty(&elemOutputId, "") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemOutput, elemOutputId) != 0) {
            proputil_property_recursive_free(elemOutputId);
            goto cleanup_get_opc_scene_config;
        }

        // dontcare
        if (proputil_create_property_empty(&elemChannelDontcare, "dontCare") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemOutputId, elemChannelDontcare) != 0) {
            proputil_property_recursive_free(elemChannelDontcare);
            goto cleanup_get_opc_scene_config;
        }

        // IgnoreLocalPriority
        if (proputil_create_property_empty(&elemIgnoreLocalPriority, "ignoreLocalPriority") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemIgnoreLocalPriority) != 0 ) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_get_opc_scene_config;
        }

        if (proputil_create_property_empty(&elemSceneDontCare, "dontCare") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemSceneDontCare) != 0) {
            proputil_property_recursive_free(elemSceneDontCare);
            goto cleanup_get_opc_scene_config;
        }

        if (proputil_create_property_empty(&elemTransitionEffect, "effect") != 0) {
            goto cleanup_get_opc_scene_config;
        }
        if (proputil_add_property_element(elemSceneId, elemTransitionEffect) != 0) {
            proputil_property_recursive_free(elemTransitionEffect);
            goto cleanup_get_opc_scene_config;
        }

        cache->scene = OPCscene;
        cache->operation = OPCoperation;

        if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_device_dsuid(dsuid, msg, cache)) {
          // free message only, if it was not injected in vdc_api_send_message
          // otherwise message will be auto removed after timeout.
          vdc_api_free_message(msg);
          free(cache);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device get opc scene). device: %d "
                            "not found\n", device_id);
    }
    return;

cleanup_get_opc_scene_config:
    free(cache);
    free(msg);
    proputil_get_property_free(submsg);
}

/*
 * Bank 1
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
static bool vdc_api_device_set_bank_1(uint16_t device_id, uint8_t param_offset, uint8_t val)
{
    char dsuid[DSUID_STR_LEN];
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;
    bool success = FALSE;

    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid))
    {
        if ((param_offset >= SensorTableStart) &&
            (param_offset < SensorTableLast))
        {
            uint8_t sensor_index, type, broadcast, pushConvert;
            uint32_t pollInterval;
            sensor_index = (param_offset - SensorTableStart);
            if (Datamodel_GetDeviceSensor(device_id, sensor_index, &type,
                                          &pollInterval, &broadcast, &pushConvert)) {
                if (sensor_index % 2) {
                    /* pollInterval can not be written through parameter writes */
                    broadcast = val & 0x08? 1 : 0;
                    pushConvert = val & 0x10? 1 : 0;
                } else {
                    type = val;
                }
                Datamodel_SetDeviceSensor(device_id, sensor_index, type,
                                          pollInterval, broadcast, pushConvert);
            } else {
                vdsm_report(LOG_DEBUG, "set sensor: no sensor found for device"
                       "%d and index %d\n", device_id, sensor_index);
            }
            return TRUE;

        } else if ((param_offset >= BinaryInputTableStart) &&
                   (param_offset < BinaryInputTableLast))
        {
            uint8_t targetGroup, type, buttonId, binInpIndex, binInpOffset;
            binInpIndex = ((param_offset - BinaryInputTableStart) / 3);
            binInpOffset = ((param_offset - BinaryInputTableStart) % 3);
            if (Datamodel_GetDeviceBinaryInput(device_id, binInpIndex,
                                               &targetGroup, &type, &buttonId)) {
                if (binInpOffset == 0) {
                    targetGroup = val;
                } else if (binInpOffset == 1) {
                    type = val;
                } else {
                    buttonId = (val >> 4);
                }
                Datamodel_SetDeviceBinaryInput(device_id, binInpIndex, targetGroup, type, buttonId);

                //--------------------------------------------------------------
                msg = malloc(sizeof(Vdcapi__Message));
                if (!msg) {
                    vdsm_report(LOG_ERR, "vdc api request set property: failed to allocate "
                                         "memory\n");
                    return FALSE;
                }
                vdcapi__message__init(msg);
                if (proputil_set_property_new(dsuid, &submsg) != 0) {
                    goto cleanup_device_set_bank_1;
                }
                msg->vdsm_request_set_property = submsg;
                msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

                // sensor function
                Vdcapi__PropertyElement *elemScene = NULL;
                Vdcapi__PropertyElement *elemGroup = NULL;
                Vdcapi__PropertyElement *inputIndex = NULL;
                Vdcapi__PropertyElement *elemSensorFunc = NULL;

                // binary input settings
                if (proputil_create_property_empty(&elemScene, "binaryInputSettings") != 0) {
                    goto cleanup_device_set_bank_1;
                }
                if (proputil_set_property_add(submsg, elemScene) != 0) {
                    proputil_property_recursive_free(elemScene);
                    goto cleanup_device_set_bank_1;
                }

                // index
                if (proputil_create_property_empty_index(&inputIndex, binInpIndex) != 0) {
                    goto cleanup_device_set_bank_1;
                }
                if (proputil_add_property_element(elemScene, inputIndex) != 0) {
                    proputil_property_recursive_free(inputIndex);
                    goto cleanup_device_set_bank_1;
                }

                // dS group number
                if (proputil_create_property_uint(&elemGroup, "group", targetGroup) != 0) {
                    goto cleanup_device_set_bank_1;
                }
                if (proputil_add_property_element(inputIndex, elemGroup) != 0) {
                    proputil_property_recursive_free(elemGroup);
                    goto cleanup_device_set_bank_1;
                }

                // sensor function
                if (proputil_create_property_uint(&elemSensorFunc, "sensorFunction", type) != 0) {
                    goto cleanup_device_set_bank_1;
                }
                if (proputil_add_property_element(inputIndex, elemSensorFunc) != 0) {
                    proputil_property_recursive_free(elemSensorFunc);
                    goto cleanup_device_set_bank_1;
                }

                // message will be freed when response is received
                int retVal =vdc_api_send_message_by_device_dsuid(dsuid, msg, NULL);
                if (VDC_API_OK == retVal) {
                    success = TRUE;
                } else if (VDC_API_ERR_CLIENT_NOT_FOUND == retVal) {
                    // free message only, if it was not injected in vdc_api_send_message
                    // otherwise message will be auto removed after timeout.
                    vdc_api_free_message(msg);
                }
            } else {
                vdsm_report(LOG_DEBUG, "set binary input: no binary input found for device %d and index %d\n", device_id, binInpIndex);
            }
        } else {
            vdsm_report(LOG_INFO, "vdc api (device set register 1)." \
                             "param offset not supported: offset:%i\n", param_offset);
        }
    } else {
        vdsm_report(LOG_WARNING, "vdc api (device set register 1). device: %d not "
                            "found\n", device_id);
    }
    return success;

cleanup_device_set_bank_1:
    free(msg);
    proputil_set_property_free(submsg);
    return FALSE;
}

/*
 * Bank 3
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
static bool vdc_api_device_set_bank_3(uint16_t device_id, uint8_t param_offset, uint8_t val)
{
    char dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client = NULL;
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;
    bool success = FALSE;

    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid) &&
        vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client))
    {
        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
            vdsm_report(LOG_ERR, "vdc api request set property: failed to allocate "
                            "memory\n");
            return FALSE;
        }
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
            goto cleanup_device_set_bank_3;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        // bank 3 offset 0 (mode)
        if (param_offset == MODE) {
            Vdcapi__PropertyElement *elemOutputSettings = NULL;
            Vdcapi__PropertyElement *elemMode = NULL;

            uint64_t mode = PRP_OUT_SETTING_MODE_DISABLE;
            switch (val) {
              case 16:
                mode = PRP_OUT_SETTING_MODE_BINARY;
              break;
              case 17:
              case 18:
              case 19:
              case 20:
              case 21:
              case 22:
              case 23:
              case 24:
              case 28:
              case 30:
              case 42:
              case 48:
              case 49:
              case 50:
              case 51:
                mode = PRP_OUT_SETTING_MODE_GRADUAL;
                break;
              case 254:
                mode = PRP_OUT_SETTING_MODE_GENERIC_ACTIVE;
                break;
            default:
                mode = PRP_OUT_SETTING_MODE_DISABLE;
            }

            // OutputSettings
            if (proputil_create_property_empty(&elemOutputSettings, "outputSettings") != 0) {
                goto cleanup_device_set_bank_3;
            }
            if (proputil_set_property_add(submsg, elemOutputSettings) != 0) {
                proputil_property_recursive_free(elemOutputSettings);
                goto cleanup_device_set_bank_3;
            }

            // Mode
            if (proputil_create_property_uint(&elemMode, "mode", mode) != 0) {
                goto cleanup_device_set_bank_3;
            }
            if (proputil_add_property_element(elemOutputSettings, elemMode) != 0) {
                proputil_property_recursive_free(elemMode);
                goto cleanup_device_set_bank_3;
            }

            // message will be freed when response is received
            if (vdc_api_send_message(vdc_client, msg, NULL) == VDC_API_OK) {
                success = TRUE;
            }

        } else if ( (param_offset == PBCHANNEL) ||
                    (param_offset == PBGROUP)   ||
                    (param_offset == LTMODE)   ||
                    (param_offset == BUTTON_ID_GRP_0)) {

            Vdcapi__PropertyElement *buttonInputSettings = NULL;
            Vdcapi__PropertyElement *elemIndex = NULL;

            // ButtonInputSettings
            if (proputil_create_property_empty(&buttonInputSettings, "buttonInputSettings") != 0) {
                goto cleanup_device_set_bank_3;
            }
            if (proputil_set_property_add(submsg, buttonInputSettings) != 0) {
                proputil_property_recursive_free(buttonInputSettings);
                goto cleanup_device_set_bank_3;
            }

            // Index: channel 0 only for this request type
            if (proputil_create_property_empty_index(&elemIndex, 0) != 0) {
                goto cleanup_device_set_bank_3;
            }
            if (proputil_add_property_element(buttonInputSettings, elemIndex)!= 0) {
                proputil_property_recursive_free(elemIndex);
                goto cleanup_device_set_bank_3;
            }

            // bank 3 offset 1 (ltnumgrp)
            if (param_offset == BUTTON_ID_GRP_0) {
                // Function
                uint64_t function = val & 0x0f;
                uint64_t group = (val>>4) & 0x0f;
                Vdcapi__PropertyElement *elemFunction = NULL;
                Vdcapi__PropertyElement *elemGroup = NULL;
                if (proputil_create_property_uint(&elemFunction, "function", function) != 0) {
                    goto cleanup_device_set_bank_3;
                }
                if (proputil_add_property_element(elemIndex, elemFunction) != 0) {
                    proputil_property_recursive_free(elemFunction);
                    goto cleanup_device_set_bank_3;
                }

                if (proputil_create_property_uint(&elemGroup, "group", group) != 0) {
                    goto cleanup_device_set_bank_3;
                }

                if (proputil_add_property_element(elemIndex, elemGroup) != 0) {
                    proputil_property_recursive_free(elemGroup);
                    goto cleanup_device_set_bank_3;
                }

            // bank 3 offset 28 (pbchannel)
            } else if (param_offset == PBCHANNEL) {
                Vdcapi__PropertyElement *elemPbChannel = NULL;

                if (proputil_create_property_uint(&elemPbChannel, "channel", val) != 0) {
                    goto cleanup_device_set_bank_3;
                }
                if (proputil_add_property_element(elemIndex, elemPbChannel) != 0) {
                    proputil_property_recursive_free(elemPbChannel);
                    goto cleanup_device_set_bank_3;
                }

            // bank 3 offset 29 (pbgroup)
            } else if (param_offset == PBGROUP) {
                Vdcapi__PropertyElement *elemPbGroup = NULL;

                if (proputil_create_property_uint(&elemPbGroup, "group", val) != 0) {
                    goto cleanup_device_set_bank_3;
                }
                if (proputil_add_property_element(elemIndex, elemPbGroup) != 0) {
                    proputil_property_recursive_free(elemPbGroup);
                    goto cleanup_device_set_bank_3;
                }

            // bank 3 offset 30 (ltmode)
            } else if (param_offset == LTMODE) {
                Vdcapi__PropertyElement *elemMode = NULL;

                if (proputil_create_property_uint(&elemMode, "mode", val) != 0) {
                    goto cleanup_device_set_bank_3;
                }
                if (proputil_add_property_element(elemIndex, elemMode) != 0) {
                    proputil_property_recursive_free(elemMode);
                    goto cleanup_device_set_bank_3;
                }
            }

            // message will be freed when response is received
            if (vdc_api_send_message(vdc_client, msg, NULL) == VDC_API_OK) {
                success = TRUE;
            }

        } else {
            vdsm_report(LOG_INFO, "vdc api (device set register 3)." \
                             "param offset not supported: offset:%i\n", param_offset);
            goto cleanup_device_set_bank_3_silent;
        }
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_WARNING, "vdc api (device set register 3). device: %d not "
                            "found\n", device_id);
    }
    return success;

cleanup_device_set_bank_3:
    vdsm_report(LOG_ERR, "vdc api request set property: failed to allocate memory\n");
cleanup_device_set_bank_3_silent:
    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    free(msg);
    proputil_set_property_free(submsg);
    return FALSE;
}

/*
 * Bank 128
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
static bool vdc_api_device_set_bank_128(uint16_t device_id, uint8_t param_offset, uint8_t val)
{
    char dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client = NULL;
    Vdcapi__Message *msg = NULL;
    Vdcapi__VdsmRequestSetProperty *submsg = NULL;
    Vdcapi__PropertyElement *elemScenes = NULL;
    Vdcapi__PropertyElement *elemSceneId = NULL;
    Vdcapi__PropertyElement *elemOutputs = NULL;
    Vdcapi__PropertyElement *elemOutputId = NULL;
    bool success = FALSE;

    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid) &&
        vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client))
    {
        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
            vdsm_report(LOG_ERR, "vdc api request set property: failed to allocate "
                            "memory\n");
            return FALSE;
        }
        vdcapi__message__init(msg);

        if (proputil_set_property_new(dsuid, &submsg) != 0) {
            goto cleanup_device_set_bank_128;
        }
        msg->vdsm_request_set_property = submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

        // Scenes
        if (proputil_create_property_empty(&elemScenes, "scenes") != 0) {
            goto cleanup_device_set_bank_128;
        }
        if (proputil_set_property_add(submsg, elemScenes) != 0) {
            proputil_property_recursive_free(elemScenes);
            goto cleanup_device_set_bank_128;
        }

        // Scene id
        int index = (param_offset & 0x7F);
        if (proputil_create_property_empty_index(&elemSceneId, index) != 0) {
            goto cleanup_device_set_bank_128;
        }
        if (proputil_add_property_element(elemScenes, elemSceneId) != 0) {
            proputil_property_recursive_free(elemSceneId);
            goto cleanup_device_set_bank_128;
        }

        // Outputs
        if (proputil_create_property_empty(&elemOutputs, "channels") != 0) {
            goto cleanup_device_set_bank_128;
        }
        if (proputil_add_property_element(elemSceneId, elemOutputs) != 0) {
            proputil_property_recursive_free(elemOutputs);
            goto cleanup_device_set_bank_128;
        }

        // Output Id
        if (proputil_create_property_empty_index(&elemOutputId, 0) != 0) {
            goto cleanup_device_set_bank_128;
        }
        if (proputil_add_property_element(elemOutputs, elemOutputId) != 0) {
            proputil_property_recursive_free(elemOutputId);
            goto cleanup_device_set_bank_128;
        }

        // value (Conversion not needed)
        if (param_offset < 128) {
            Vdcapi__PropertyElement *elemValue = NULL;

            if (proputil_create_property_double(&elemValue, "value", val) != 0) {
                goto cleanup_device_set_bank_128;
            }
            if (proputil_add_property_element(elemOutputId, elemValue) != 0) {
                proputil_property_recursive_free(elemValue);
                goto cleanup_device_set_bank_128;
            }
        } else {
            // configuration
            bool dontcare = (val & 0x01);
            bool ignoreLocalPriority = (val & 0x02);
            bool flashmode = (val & 0x08);

            uint64_t dimtime = (val >> 6) & 0x3; // 1:standard, 2:slow, 3:unknown, 0:should not occur -> map to standard
            if (dimtime==0) {
                // dimtime==0 is a terminal block internal "don't change dimtime" value that should not occur externally
                dimtime = 1; // use standard dimtime instead
            }

            // effect 4 is "blink/alert" in vDC API
            // effect 1..3 correspond to dimtime 1..3
            uint64_t effect = flashmode ? 4 : dimtime;

            Vdcapi__PropertyElement *elemDontCare = NULL;
            Vdcapi__PropertyElement *elemIgnoreLocalPriority = NULL;
            Vdcapi__PropertyElement *elemEffect = NULL;

            // dontCare
            if (proputil_create_property_bool(&elemDontCare, "dontCare", dontcare) != 0) {
                goto cleanup_device_set_bank_128;
            }
            if (proputil_add_property_element(elemSceneId, elemDontCare) != 0) {
                proputil_property_recursive_free(elemDontCare);
                goto cleanup_device_set_bank_128;
            }

            // ignoreLocalPriority
            if (proputil_create_property_bool(&elemIgnoreLocalPriority, "ignoreLocalPriority", ignoreLocalPriority) != 0) {
                goto cleanup_device_set_bank_128;
            }
            if (proputil_add_property_element(elemSceneId, elemIgnoreLocalPriority) != 0) {
                proputil_property_recursive_free(elemIgnoreLocalPriority);
                goto cleanup_device_set_bank_128;
            }

            if (proputil_create_property_uint(&elemEffect, "effect", effect) != 0) {
                goto cleanup_device_set_bank_128;
            }
            if (proputil_add_property_element(elemSceneId, elemEffect) != 0) {
                proputil_property_recursive_free(elemEffect);
                goto cleanup_device_set_bank_128;
            }
        }

        // message will be freed when response is received
        if (vdc_api_send_message(vdc_client, msg, NULL) == VDC_API_OK) {
            success = TRUE;
        }
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_WARNING, "vdc api (device set register 128). device: %d not "
                            "found\n", device_id);
    }
    return success;

cleanup_device_set_bank_128:
    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    free(msg);
    proputil_set_property_free(submsg);
    return FALSE;
}

/*
 * Banks
 * https://intranet.aizo.net/Wiki/Parametersatz.aspx
 */
bool vdc_api_device_set_register(uint16_t device_id, uint8_t param_bank,
                                 uint8_t param_offset, uint8_t val)
{
    bool success = FALSE;
    switch (param_bank) {
    case 1:   success = vdc_api_device_set_bank_1(device_id, param_offset, val); break;
    case 3:   success = vdc_api_device_set_bank_3(device_id, param_offset, val); break;
    case 103: success = vdc_api_device_set_opc_scene_config(device_id, param_offset, val); break;
    case 104: success = vdc_api_device_set_opc_scene_dc_flags(device_id, param_offset, val); break;
    case 128: success = vdc_api_device_set_bank_128(device_id, param_offset, val); break;
    default:
        vdsm_report(LOG_WARNING, "vdc api (device set register). offset %d on "
                            "bank %d not handled\n", param_offset, param_bank);
    }
    return success;
}

void vdc_api_device_get_sensor_value(uint16_t device_id, uint8_t sensor_index)
{
    char dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client;

    vdsm_report(LOG_DEBUG, "vdc api get device sensor value. device %d, sensor %d\n",
           device_id, sensor_index);

    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (Datamodel_GetDsuidFromDeviceId(device_id, dsuid) &&
        vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client))
    {
        request_cache_t *cache;
        cache = (request_cache_t*) malloc(sizeof(request_cache_t));
        if (cache == NULL) {
            pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
            vdsm_report(LOG_ERR, "could not allocate memory for property request "
                            "cache\n");
            return;
        }

        cache->context = REQUEST_CONTEXT_GET_PROPERTY_SENSOR_VALUE;
        cache->channel = sensor_index;
        cache->result = device_id;

        if (sensor_index == 0x20) {
            vdc_api_send_get_property(vdc_client, dsuid, "binaryInputStates",
                                      cache);

// vDC-API does not support the Device status property yet.
// This snippet serves as a reminder, that the vDC-API needs to be extended.
//        } else if (sensor_index >= 0x10 && sensor_index < 0x20) {
//            uint32_t data[1];
//            data[0] = sensor_index;
//            vdc_api_send_get_properties_indexed(vdc_client, dsuid, "deviceStatus",
//                                                data, 1, cache);

        } else {
            uint32_t data[1];
            data[0] = sensor_index;
            vdc_api_send_get_properties_indexed(vdc_client, dsuid, "sensorStates",
                                                data, 1, cache);
        }
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_WARNING, "vdc api get device sensor value. device: %d not "
                            "found\n", device_id);
    }
}

void vdc_api_device_get_zone(const char* dsuid)
{
    vdsm_report(LOG_DEBUG, "vdc api get device zone. device %s\n", dsuid);
    request_cache_t *cache;
    cache = (request_cache_t*) malloc(sizeof(request_cache_t));
    if (cache == NULL)
    {
        vdsm_report(LOG_ERR, "could not allocate memory for "
                        "property request cache\n");
        return;
    }
    cache->context = REQUEST_CONTEXT_GET_PROPERTY_ZONEID;
    vdc_api_send_get_property_by_device_dsuid(dsuid, "zoneID", cache);
}

void vdc_api_zone_push_sensor_value(uint8_t localZoneId, uint8_t groupId,
                                    uint8_t sensorType, uint16_t sensorValue)
{
    device_dsuid_t *dsuid_list = NULL, *dsuid, *tmp_dsuid;
    vdc_client_t *vdc_client;

    vdsm_report(LOG_DEBUG, "vdc api push sensor value to local zone %d, sensor "
                      "type %d, sensor value %d\n",
                      localZoneId, sensorType, sensorValue);
    Datamodel_ZoneGroupGetDeviceDsuids(localZoneId, groupId, &dsuid_list);

    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    LL_FOREACH_SAFE(dsuid_list, dsuid, tmp_dsuid) {
        if (vdc_api_get_vdc_from_lookup_table_private(dsuid->dsuid, &vdc_client)) {
            vdsm_report(LOG_DEBUG, "vdc api push sensor value to device %s\n",
                   dsuid->dsuid);

            Vdcapi__Message msg = VDCAPI__MESSAGE__INIT;
            Vdcapi__VdsmNotificationSetControlValue submsg =
                VDCAPI__VDSM__NOTIFICATION_SET_CONTROL_VALUE__INIT;

            char *dsuids[1];
            dsuids[0] = dsuid->dsuid;
            submsg.dsuid = dsuids;
            submsg.n_dsuid = 1;

            switch (sensorType) {
                case 9:
                    submsg.value = ((double)sensorValue * 0.1 / 4) - 273.15 + 230;
                    submsg.has_value = 1;
                    submsg.name = "TemperatureZone";
                    break;
                case 10:
                    submsg.value = ((double)sensorValue * 0.1 / 4) - 273.15 + 230;
                    submsg.has_value = 1;
                    submsg.name = "TemperatureOutside";
                    break;
                case 11:
                    submsg.value = pow(10, (double)sensorValue / 800);
                    submsg.has_value = 1;
                    submsg.name = "BrightnessZone";
                    break;
                case 12:
                    submsg.value = pow(10, (double)sensorValue / 800);
                    submsg.has_value = 1;
                    submsg.name = "BrightnessOutside";
                    break;
                case 13:
                    submsg.value = ((double)sensorValue * 0.1 / 4);
                    submsg.has_value = 1;
                    submsg.name = "HumidityZone";
                    break;
                case 14:
                    submsg.value = ((double)sensorValue * 0.1 / 4);
                    submsg.has_value = 1;
                    submsg.name = "HumidityOutside";
                    break;
                case 15:
                    submsg.value = ((double)sensorValue + 200.0);
                    submsg.has_value = 1;
                    submsg.name = "AirPressure";
                    break;
                case 18:
                    submsg.value = ((double)sensorValue * 0.1 / 4);
                    submsg.has_value = 1;
                    submsg.name = "WindSpeed";
                    break;
                case 19:
                    submsg.value = ((double)sensorValue);
                    submsg.has_value = 1;
                    submsg.name = "WindDirection";
                    break;
                case 20:
                    submsg.value = ((double)sensorValue);
                    submsg.has_value = 1;
                    submsg.name = "Precipitation";
                    break;
                case 21:
                    submsg.value = pow(10, (double)sensorValue / 800);
                    submsg.has_value = 1;
                    submsg.name = "CarbonDioxide";
                    break;
                case 25:
                    submsg.value = ((double)sensorValue * 0.25 / 4);
                    submsg.has_value = 1;
                    submsg.name = "SoundPressureLevel";
                    break;
                case 50:
                    submsg.value = ((double)sensorValue * 0.1 / 4) - 273.15 + 230;
                    submsg.has_value = 1;
                    submsg.name = "TemperatureSetPoint";
                    break;
                case 51:
                    submsg.value = (double)sensorValue - 100.0;
                    submsg.has_value = 1;
                    submsg.name = "heatingLevel";
                    break;
                default:
                    vdsm_report(LOG_INFO, "vdc api push sensor value. sensor "
                                     "type %d not handled \n", sensorType);
                    goto freeDevice;
            }

            msg.type = VDCAPI__TYPE__VDSM_NOTIFICATION_SET_CONTROL_VALUE;
            msg.vdsm_send_set_control_value = &submsg;
            vdc_api_send_message(vdc_client, &msg, NULL);
        } else {
            vdsm_report(LOG_WARNING, "vdc api push sensor value. device: %s not "
                                "found in lookup table\n", dsuid->dsuid);
        }
freeDevice:
        LL_DELETE(dsuid_list, dsuid);
        free(dsuid);
    }
    pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
}

bool vdc_api_vdsm_holds_vdc(const dsuid_t *vdc_dsuid)
{
    bool success = false;
    char destAddr [DSUID_STR_LEN];
    dsuid_to_string_upper(vdc_dsuid, destAddr);

    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);

    vdc_lookup_hash_t *vdc_lookup;
    HASH_FIND_STR(g_vdc_lookup_hash, destAddr, vdc_lookup);
    if (vdc_lookup != NULL) {
        success = true;
    } else {
        /* check if vdsm */
        if (dsuid_equal(&g_dsm_dsuid, vdc_dsuid)) {
            vdsm_report(LOG_DEBUG, "vdc api addressing the vdsm %s\n", destAddr);
            success  = true;
        } else {
            vdsm_report(LOG_WARNING, "vdc api vdsm hold vdc not found %s\n", destAddr);
        }
    }
    pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);

    return success;
}

void vdc_api_device_send_announce_ok(uint32_t messageId, uint16_t deviceId)
{
    char dsuid[DSUID_STR_LEN];
    vdc_client_t *vdc_client;
    pthread_mutex_lock(&g_vdc_device_lookup_hash_mutex);
    if (Datamodel_GetDsuidFromDeviceId(deviceId, dsuid) &&
        vdc_api_get_vdc_from_lookup_table_private(dsuid, &vdc_client)) {
        vdc_api_send_ok_response(vdc_client, messageId);
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_device_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "could not send back announce ok message for device:\n", deviceId);
    }
}

void vdc_api_vdc_send_announce_ok(uint32_t messageId, const char* vdc_dsuid)
{
    // resolve vdc_dsuid to vdc_client
    vdc_lookup_hash_t *vdc_lookup;
    pthread_mutex_lock(&g_vdc_lookup_hash_mutex);
    HASH_FIND_STR(g_vdc_lookup_hash, vdc_dsuid, vdc_lookup);
    if (vdc_lookup != NULL) {
        vdc_api_send_ok_response(vdc_lookup->vdc_client, messageId);
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
    } else {
        pthread_mutex_unlock(&g_vdc_lookup_hash_mutex);
        vdsm_report(LOG_ERR, "could not send back announce ok message for vdc: %s\n", vdc_dsuid);
    }
}

void vdc_api_vdc_get_name(const char *vdc_str)
{
    request_cache_t *cache = NULL;

    cache = (request_cache_t*) malloc(sizeof(request_cache_t));
    if (!cache) {
        vdsm_report(LOG_ERR, "vdc api: vdc get name: could not allocate memory "
                             "for property request cache\n");
        return;
    }

    cache->context = REQUEST_CONTEXT_GET_VDC_NAME;
    vdc_api_send_get_property_by_vdc_dsuid(vdc_str, "name", cache);
}

void vdc_api_vdc_set_name(const dsuid_t *vdcdsuid, const char *name)
{
    char vdc_str[DSUID_STR_LEN];
    dsuid_to_string_upper(vdcdsuid, vdc_str);

    Vdcapi__Message *msg;
    Vdcapi__VdsmRequestSetProperty *submsg;

    msg = malloc(sizeof(Vdcapi__Message));
    if (!msg) {
        vdsm_report(LOG_ERR, "vdc api vdc set name: failed to allocate "
                    "memory for message body\n");
        return;
    }
    vdcapi__message__init(msg);

    if (proputil_set_property_new(vdc_str, &submsg) != 0) {
        vdsm_report(LOG_ERR,"vdc api: vdc set name: failed to set property dsuid\n");
        free(msg);
        return;
    }
    msg->vdsm_request_set_property = submsg;
    msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

    Vdcapi__PropertyElement* propElem = NULL;
    if (proputil_set_property_add_string(submsg, "name", name, &propElem) != 0) {
        vdsm_report(LOG_ERR,"vdc api: vdc set name: failed to set property name\n");
        free(msg);
        proputil_set_property_free(submsg);
        return;
    }

    // message will be freed when response is received
    if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_vdc_dsuid(vdc_str, msg, NULL)) {
        vdsm_report(LOG_ERR,
            "FAILED: vdc api: vdc set name for vdc  %s\n", vdc_str);

        // free message only, if it was not injected in vdc_api_send_message
        // otherwise message will be auto removed after timeout.
        vdc_api_free_message(msg);
    }
}

void vdc_api_vdc_set_defaultZoneID(const char *vdc_str, uint16_t zoneID)
{
    Vdcapi__Message *msg;
    Vdcapi__VdsmRequestSetProperty *submsg;

    msg = malloc(sizeof(Vdcapi__Message));
    if (!msg) {
        vdsm_report(LOG_ERR, "vdc api vdc set default zone id: failed to allocate "
                    "memory for message body\n");
        return;
    }
    vdcapi__message__init(msg);

    if (proputil_set_property_new(vdc_str, &submsg) != 0) {
        vdsm_report(LOG_ERR,"vdc api: vdc set zone id: failed to set property dsuid\n");
        free(msg);
        return;
    }
    msg->vdsm_request_set_property = submsg;
    msg->type = VDCAPI__TYPE__VDSM_REQUEST_SET_PROPERTY;

    if (proputil_set_property_add_uint(submsg, "zoneID", zoneID) != 0) {
        vdsm_report(LOG_ERR,"vdc api: vdc set zoneID: failed to set property name\n");
        free(msg);
        proputil_set_property_free(submsg);
        return;
    }

    // message will be freed when response is received
    if (VDC_API_ERR_CLIENT_NOT_FOUND == vdc_api_send_message_by_vdc_dsuid(vdc_str, msg, NULL)) {
        vdsm_report(LOG_ERR,
            "FAILED: vdc api: vdc set name for vdc  %s\n", vdc_str);

        // free message only, if it was not injected in vdc_api_send_message
        // otherwise message will be auto removed after timeout.
        vdc_api_free_message(msg);
    }
}

void vdc_api_vdc_get_defaultZoneID(const char *vdc_str)
{
    request_cache_t *cache = NULL;

    cache = (request_cache_t*) malloc(sizeof(request_cache_t));
    if (!cache) {
        vdsm_report(LOG_ERR, "vdc api: vdc get default zone id: could not allocate memory "
                             "for property request cache\n");
        return;
    }

    cache->context = REQUEST_CONTEXT_GET_VDC_ZONDEID;
    vdc_api_send_get_property_by_vdc_dsuid(vdc_str, "zoneID", cache);
}

void vdc_api_vdc_get_capabilities(const char *vdc_str)
{
    request_cache_t *cache = NULL;

    cache = (request_cache_t*) malloc(sizeof(request_cache_t));
    if (!cache) {
        vdsm_report(LOG_ERR, "vdc api: vdc get capabilities: could not allocate memory "
                             "for property request cache\n");
        return;
    }

    cache->context = REQUEST_CONTEXT_GET_VDC_CAPABILITIES;
    vdc_api_send_get_property_by_vdc_dsuid(vdc_str, "capabilities", cache);
}

device_dsuid_t *singleDsuidList(char *dsuid)
{
    device_dsuid_t *l = (device_dsuid_t*) malloc(sizeof(device_dsuid_t));
    if (!l) {
        vdsm_report(LOG_ERR,
            "FAILED: singleDsuidList: can not allocate memory  for  device %s\n", dsuid);
        return NULL;
    }
    strcpy(l->dsuid, dsuid);
    l->next = NULL;
    return l;
}
