/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Remy Mahler, digitalSTROM AG <remy.mahler@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif

#include "vdcProcessor.h"
#include "bsp.h"
#include "util.h"
#include "vdc-api-proto.h"

Q_DEFINE_THIS_MODULE(vdcProcessor)

enum {
    VDC_WAIT_FOR_DSID_TICKS            = BSP_TICKS(5000)
};

typedef struct {
    QEvent super;
    char vdcdsuid[DSUID_STR_LEN];
    uint32_t messageId;
} VDCAnnounceVDCEvt;

/* Active object class -----------------------------------------------------*/
typedef struct {
    QActive super;                          /**< active object base class */
    QTimeEvt timeEvt;                       /**< timer */
    char vdc_dsuid[DSUID_STR_LEN];          /**< selected vdc */
    uint32_t messageId;                     /**< vdc message id */
    QEQueue deferredCommandQueue;           /**< deferred command event queue */
} VdcProcessor;

/* Local objects -----------------------------------------------------------*/
static VdcProcessor l_vdcProcessor;
static QEvent const * l_deferredCommandQueueSto[12];/* deferred queue storage */

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_vdcProcessor = (QActive *)&l_vdcProcessor;/* "opaque" AO pointer */

/* Local methods  forward declaration --------------------------------------*/
static QState VDC_initial(VdcProcessor *me, QEvent const *e);

static QState VDC_idle(VdcProcessor *me, QEvent const *e);
static QState VDC_busy(VdcProcessor *me, QEvent const *e);

static QState VDC_get_name(VdcProcessor *me, QEvent const *e);
static QState VDC_get_ZoneID(VdcProcessor *me, QEvent const *e);
static QState VDC_get_capability(VdcProcessor *me, QEvent const *e);

/*..........................................................................*/
void vdcProcessor_ctor(void)
{
    VdcProcessor *me = &l_vdcProcessor;
    /* call active object (superclass) constructor */
    QActive_ctor(&me->super, (QStateHandler)&VDC_initial);
    /* initialize time events */
    QTimeEvt_ctor(&me->timeEvt, VDC_TIMEOUT_SIG);
    QEQueue_init(&me->deferredCommandQueue, l_deferredCommandQueueSto, Q_DIM(l_deferredCommandQueueSto));
}

/*..........................................................................*/
static QState VDC_initial(VdcProcessor *me, QEvent const *e __attribute__((unused)))
{
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_vdcProcessor);

    /* QS signal dictionary */
    QS_SIG_DICTIONARY(VDC_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(VDC_UPSTREAM_SIG, 0);
    QS_SIG_DICTIONARY(VDC_ANNOUNCE_VDC_SIG, 0);

    /* QS function dictionary */
    QS_FUN_DICTIONARY(&VDC_initial);
    QS_FUN_DICTIONARY(&VDC_idle);
    QS_FUN_DICTIONARY(&VDC_busy);
    QS_FUN_DICTIONARY(&VDC_get_name);
    QS_FUN_DICTIONARY(&VDC_get_ZoneID);
    QS_FUN_DICTIONARY(&VDC_get_capability);

    /* QS subscribe events */
    QActive_subscribe((QActive *)me, VDC_UPSTREAM_SIG);

    return Q_TRAN(&VDC_idle);
}

/*..........................................................................*/
static QState VDC_idle(VdcProcessor *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_idle Q_ENTRY_SIG\n");
        /* recall deferred command event */
        if (QActive_recall((QActive *)me, &me->deferredCommandQueue) != 0) {
            vdsm_report(LOG_INFO, "vdc stm process deferred messages\n");
        }
        return Q_HANDLED();
    }
    case VDC_ANNOUNCE_VDC_SIG: {
        VDCAnnounceVDCEvt* evt = (VDCAnnounceVDCEvt *)e;
        strncpy(me->vdc_dsuid, evt->vdcdsuid, sizeof(me->vdc_dsuid));
        me->messageId = evt->messageId;
        vdsm_report(LOG_INFO, "VDC_processor start processing vdc: %s\n", me->vdc_dsuid);
        return Q_TRAN(&VDC_get_name);
    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState VDC_busy(VdcProcessor *me, QEvent const *e)
{
    switch (e->sig) {
    case VDC_ANNOUNCE_VDC_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_busy VDC_ANNOUNCE_VDC_SIG\n");
        if (QEQueue_getNFree(&me->deferredCommandQueue) != 0) {
            /* defer command */
            QActive_defer((QActive *)me, &me->deferredCommandQueue, e);
        } else {
            vdsm_report(LOG_ERR, "VDC_busy throwing away message\n");
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState VDC_get_name(VdcProcessor *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_name Q_ENTRY_SIG\n");
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, VDC_WAIT_FOR_DSID_TICKS);
        vdc_api_vdc_get_name(me->vdc_dsuid);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case VDC_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT VDC_get_name\n");
        return Q_TRAN(&VDC_idle);
    }
    case VDC_UPSTREAM_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_name VDC_UPSTREAM_SIG\n");
        /* already in database */
        /* LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e; */
        return Q_TRAN(&VDC_get_ZoneID);
    }
    }
    return Q_SUPER(&VDC_busy);
}

/*..........................................................................*/
static QState VDC_get_ZoneID(VdcProcessor *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_ZoneID Q_ENTRY_SIG\n");
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, VDC_WAIT_FOR_DSID_TICKS);
        vdc_api_vdc_get_defaultZoneID(me->vdc_dsuid);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case VDC_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT VDC_get_zoneID\n");
        return Q_TRAN(&VDC_idle);
    }
    case VDC_UPSTREAM_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_ZoneID VDC_UPSTREAM_SIG\n");
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        uint16_t zoneId = longUpEvt->data;
        if (zoneId > 0) {
            Datamodel_setVdcDefaultZoneID(me->vdc_dsuid, zoneId);
        } else {
            dsuid_t vdc_dsuid;
            dsuid_from_string(me->vdc_dsuid, &vdc_dsuid);
            zoneId = Datamodel_create_zoneId_from_dsuid(&vdc_dsuid);
            Datamodel_setVdcDefaultZoneID(me->vdc_dsuid, zoneId);
            vdc_api_vdc_set_defaultZoneID(me->vdc_dsuid, zoneId);
        }
        return Q_TRAN(&VDC_get_capability);
    }
    }
    return Q_SUPER(&VDC_busy);
}

/*..........................................................................*/
static QState VDC_get_capability(VdcProcessor *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_capability Q_ENTRY_SIG\n");
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, VDC_WAIT_FOR_DSID_TICKS);
        vdc_api_vdc_get_capabilities(me->vdc_dsuid);
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_TRAN(&VDC_idle);
    }
    case VDC_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT VDC_get_capability\n");
        return Q_TRAN(&VDC_idle);
    }
    case VDC_UPSTREAM_SIG: {
        vdsm_report(LOG_DEBUG, "VDC_get_capability VDC_UPSTREAM_SIG\n");

        /* LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e; */
        /* TODO: capabilities not yet used */

        vdc_api_vdc_send_announce_ok(me->messageId, me->vdc_dsuid);
        /* announce vdc */
        DS485_EmitBusChange(me->vdc_dsuid, 0);
        vdsm_report(LOG_INFO, "VDC_processor finished processing vdc: %s\n", me->vdc_dsuid);

        return Q_TRAN(&VDC_idle);
    }
    }
    return Q_SUPER(&VDC_busy);
}

/*..........................................................................*/
void VDC_AnnounceVDC(const char *vdc_dsuid, uint32_t message_id)
{
    VDCAnnounceVDCEvt* evt = Q_NEW(VDCAnnounceVDCEvt, VDC_ANNOUNCE_VDC_SIG);
    strncpy(evt->vdcdsuid, vdc_dsuid, sizeof(evt->vdcdsuid));
    evt->messageId = message_id;
    QActive_postFIFO(AO_vdcProcessor, (QEvent *)evt);
}
